"""The misc (miscellaneous) module contains filters, which don't have a classical purpose."""
import subprocess
import os
import tempfile
import typing as t

import numpy as np
import SimpleITK as sitk

import pymia.filtering.filter as pymia_fltr


class Relabel(pymia_fltr.IFilter):
    """Relabels the labels in the file by the provided rule"""

    def __init__(self, label_changes: t.Dict[int, t.Union[int, tuple]]) -> None:
        """Initializes a new instance of the LargestNComponents class.

        Args:
            label_changes(Dict[int, Union[int, tuple]]): Label change rule where the key is the new label and
                the value the existing (can be multiple)
        """
        super().__init__()
        self.label_changes = label_changes

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams = None) -> sitk.Image:
        """Executes the relabeling of the label image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The filtered image.
        """
        np_img = sitk.GetArrayFromImage(image)
        new_np_img = np_img.copy()
        for new_label, old_labels in self.label_changes.items():
            mask = np.in1d(np_img.ravel(), old_labels).reshape(np_img.shape)
            new_np_img[mask] = new_label
        new_img = sitk.GetImageFromArray(new_np_img)
        new_img.CopyInformation(image)
        return new_img

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        str_list = []
        for k, v in self.label_changes.items():
            str_list.append('{}->{}'.format(k, v))
        return 'Relabel:\n' \
               ' label_changes: {label_changes}\n' \
            .format(self=self, label_changes='; '.join(str_list))


class SizeCorrectionParams(pymia_fltr.IFilterParams):
    """Size (shape) correction filter parameters."""

    def __init__(self, reference_shape: tuple) -> None:
        """Initializes a new instance of the SizeCorrectionParams class.

        Args:
            reference_shape (tuple): The reference or target shape.
        """
        self.dims = len(reference_shape)
        self.reference_shape = reference_shape


class SizeCorrectionFilter(pymia_fltr.IFilter):
    """A method to correct shape/size difference."""

    def __init__(self, two_sided=True, pad_constant=0.0) -> None:
        """ Initializes a new instance of the SizeCorrectionFilter class.

        Args:
            two_sided (bool): whether the cropping and padding should be applied on one or both side of the dimension.
            pad_constant (int): constant used for padding
        """
        super().__init__()
        self.two_sided = two_sided
        self.pad_constant = pad_constant

    def execute(self, image: sitk.Image, params: SizeCorrectionParams = None) -> sitk.Image:
        """Executes the shape/size correction by help of padding and cropping.

        Args:
            image (sitk.Image): The image.
            params (SizeCorrectionParams): The parameters containing the reference (target) shape.

        Returns:
            sitk.Image: The filtered image.
        """
        if params is None:
            raise ValueError('ShapeParams argument is missing')
        if image.GetDimension() != params.dims:
            raise ValueError('image dimension {} is not compatible with reference shape dimension {}'
                             .format(image.GetDimension(), params.dims))

        image_shape = image.GetSize()
        crop = [params.dims*[0], params.dims*[0]]
        pad = [params.dims*[0], params.dims*[0]]
        for dim in range(params.dims):
            ref_size = params.reference_shape[dim]
            dim_size = image_shape[dim]
            if dim_size > ref_size:
                if self.two_sided:
                    crop[0][dim] = (dim_size - ref_size) // 2
                    crop[1][dim] = (dim_size - ref_size) // 2 + ((dim_size - ref_size) % 2)
                else:
                    crop[0][dim] = (dim_size - ref_size)
            elif dim_size < ref_size:
                if self.two_sided:
                    pad[0][dim] = (ref_size - dim_size) // 2
                    pad[1][dim] = (ref_size - dim_size) // 2 + ((ref_size - dim_size) % 2)
                else:
                    pad[0][dim] = (ref_size - dim_size)

        crop_needed = any(any(c) for c in crop)
        if crop_needed:
            image = sitk.Crop(image, crop[0], crop[1])

        pad_needed = any(any(p) for p in pad)
        if pad_needed:
            image = sitk.ConstantPad(image, pad[0], pad[1], self.pad_constant)

        return image

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'SizeCorrectionFilter:\n' \
               ' two_sided: {self.two_sided}\n' \
            .format(self=self)


class CmdlineExecutorParams(pymia_fltr.IFilterParams):
    """Command line executor filter parameters."""

    def __init__(self, arguments: t.List[str]) -> None:
        """Initializes a new instance of the CmdlineExecutorParams class.

        Args:
            arguments (t.List[str]): Additional arguments for the command line execution.
        """
        self.arguments = arguments


class CmdlineExecutor(pymia_fltr.IFilter):
    """Represents a command line executable."""

    def __init__(self, executable_path: str):
        """Initializes a new instance of the CmdlineExecutor class.

        Args:
            executable_path (str): The path to the executable to run.
        """
        super().__init__()
        self.executable_path = executable_path

    def execute(self, image: sitk.Image, params: CmdlineExecutorParams=None) -> sitk.Image:
        """Executes a command line program.

        Args:
            image (SimpleITK.Image): The image.
            params (CmdlineExecutorParams): The execution specific command line parameters.

        Returns:
            sitk.Image: The filtered image.
        """
        temp_dir = tempfile.gettempdir()
        temp_in = os.path.join(temp_dir, 'in.nii')
        sitk.WriteImage(image, temp_in)
        temp_out = os.path.join(temp_dir, 'out.nii')
        
        cmd = [self.executable_path, temp_in, temp_out]
        if params is not None:
            cmd = cmd + params.arguments
        
        subprocess.run(cmd, check=True)
        out_image = sitk.ReadImage(temp_out, image.GetPixelID())

        # clean up
        os.remove(temp_in)
        os.remove(temp_out)

        return out_image

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'CmdlineExecutor:\n' \
               ' executable_path: {self.executable_path}\n' \
            .format(self=self)



from .callback import (Callback, MonitoringCallback, WriteDataCallback, WriteFilesCallback, WriteNamesCallback,
                       WriteSubjectCallback, WriteImageInformationCallback, ComposeCallback, get_default_callbacks)
from .fileloader import (Load, LoadDefault)
from .writer import (Hdf5Writer, Writer, get_writer)
from .traverser import (SubjectFileTraverser, Traverser)



import abc
import pickle
import typing

import numpy as np

import pymia.data.transformation as tfm


class Assembler(abc.ABC):

    @abc.abstractmethod
    def add_batch(self, prediction, batch: dict):
        pass

    @abc.abstractmethod
    def get_assembled_subject(self, subject_index: int):
        pass

    @property
    @abc.abstractmethod
    def subjects_ready(self):
        pass


def numpy_zeros(shape: tuple, id_: str, batch: dict, idx: int):
    return np.zeros(shape)


def default_sample_fn(params: dict):
    key = params['key']
    batch = params['batch']
    idx = params['batch_idx']

    data = params[key]
    index_expr = batch['index_expr'][idx]
    if isinstance(index_expr, bytes):
        # is pickled
        index_expr = pickle.loads(index_expr)

    return data, index_expr


class SubjectAssembler(Assembler):
    """Assembles predictions of one or multiple subjects.

    Assumes that the network output, i.e. to_assemble, is of shape (B, ..., C)
    where B is the batch size and C is the numbers of channels (must be at least 1) and ... refers to an arbitrary image
    dimension.
    """

    def __init__(self, zero_fn=numpy_zeros, on_sample_fn=default_sample_fn):
        """Initializes a new instance of the SubjectAssembler class.

        Args:
            zero_fn: A function that initializes the numpy array to hold the predictions.
                Args: shape: tuple with the shape of the subject's labels, id_: str identifying the subject.
                Returns: A np.ndarray
            on_sample_fn: A function that processes a sample.
                Args: params: dict with the parameters.
                Returns tuple of data and index expression, i.e. the processed data and index expression.
        """
        self.predictions = {}
        self._subjects_ready = set()
        self.zero_fn = zero_fn
        self.on_sample_fn = on_sample_fn

    @property
    def subjects_ready(self):
        return self._subjects_ready

    def add_batch(self, to_assemble, batch: dict, last_batch=False):
        if 'subject_index' not in batch:
            raise ValueError('SubjectAssembler requires "subject_index" to be extracted (use IndexingExtractor)')
        if 'index_expr' not in batch:
            raise ValueError('SubjectAssembler requires "index_expr" to be extracted (use IndexingExtractor)')
        if 'shape' not in batch:
            raise ValueError('SubjectAssembler requires "shape" to be extracted (use ImageShapeExtractor)')

        if not isinstance(to_assemble, dict):
            to_assemble = {'__prediction': to_assemble}

        for idx in range(len(batch['subject_index'])):
            self.add_sample(to_assemble, batch, idx)

        if last_batch:
            # to prevent from last batch to be ignored
            self.end()

    def end(self):
        self._subjects_ready = set(self.predictions.keys())

    def add_sample(self, to_assemble, batch, idx):

        subject_index = batch['subject_index'][idx]

        # initialize subject
        if subject_index not in self.predictions and not self.predictions:
            self.predictions[subject_index] = self._init_new_subject(batch, to_assemble, idx)
        elif subject_index not in self.predictions:
            self._subjects_ready = set(self.predictions.keys())
            self.predictions[subject_index] = self._init_new_subject(batch, to_assemble, idx)

        for key in to_assemble:
            data = to_assemble[key][idx]
            sample_data, index_expr = self.on_sample_fn({'key': key,
                                                         key: data,
                                                         'batch': batch,
                                                         'batch_idx': idx,
                                                         'predictions': self.predictions})

            self.predictions[subject_index][key][index_expr.expression] = sample_data

    def _init_new_subject(self, batch, to_assemble, idx):
        subject_prediction = {}
        for key in to_assemble:
            subject_shape = batch['shape'][idx]
            subject_shape += (to_assemble[key].shape[-1],)
            subject_prediction[key] = self.zero_fn(subject_shape, key, batch, idx)
        return subject_prediction

    def get_assembled_subject(self, subject_index: int):
        """Gets the prediction of a subject.

        Args:
            subject_index (int): The subject's index.

        Returns:
            np.ndarray: The prediction of the subject.
        """
        try:
            self._subjects_ready.remove(subject_index)
        except KeyError:
            # check if subject is assembled but not listed as ready
            # this can happen if only one subject was assembled or last
            if subject_index not in self.predictions:
                raise ValueError('Subject with index {} not in assembler'.format(subject_index))
        assembled = self.predictions.pop(subject_index)
        if '__prediction' in assembled:
            return assembled['__prediction']
        return assembled


def mean_merge_fn(planes: list):
    return np.stack(planes).mean(axis=0)


class TransformSampleFn:

    def __init__(self, transform: tfm.Transform) -> None:
        self.transform = transform

    def __call__(self, params: dict):
        key = params['key']
        batch = params['batch']
        idx = params['batch_idx']

        index_expr = batch['index_expr'][idx]
        if isinstance(index_expr, bytes):
            # is pickled
            index_expr = pickle.loads(index_expr)

        temp = tfm.raise_error_if_entry_not_extracted
        tfm.raise_error_entry_not_extracted = False
        ret = self.transform({key: params[key], 'index_expr': index_expr})
        tfm.raise_error_entry_not_extracted = temp

        return ret[key], ret['index_expr']


class PlaneSubjectAssembler(Assembler):
    def __init__(self, merge_fn=mean_merge_fn, zero_fn=numpy_zeros):
        self.planes = {}  # type: typing.Dict[int, SubjectAssembler]
        self._subjects_ready = set()
        self.zero_fn = zero_fn
        self.merge_fn = merge_fn

    @property
    def subjects_ready(self):
        return self._subjects_ready

    def add_batch(self, to_assemble, batch: dict, last_batch=False):
        if 'index_expr' not in batch:
            raise ValueError('SubjectAssembler requires "index_expr" to be extracted (use IndexingExtractor)')
        if 'shape' not in batch:
            raise ValueError('SubjectAssembler requires "shape" to be extracted (use ImageShapeExtractor)')

        if not isinstance(to_assemble, dict):
            to_assemble = {'__prediction': to_assemble}

        for idx in range(len(batch['index_expr'])):
            index_expr = batch['index_expr'][idx]
            if isinstance(index_expr, bytes):
                # is pickled
                index_expr = pickle.loads(index_expr)
            plane_dimension = self._get_plane_dimension(index_expr)

            if plane_dimension not in self.planes:
                self.planes[plane_dimension] = SubjectAssembler(self.zero_fn)

            indexing = index_expr.get_indexing()
            if not isinstance(indexing, list):
                indexing = [indexing]

            required_plane_shape = list(batch['shape'][idx])

            index_at_plane = indexing[plane_dimension]
            if isinstance(index_at_plane, tuple):
                # is a range in the off plane direction (tuple)
                required_off_plane_size = index_at_plane[1] - index_at_plane[0]
                required_plane_shape[plane_dimension] = required_off_plane_size
            else:  # isinstance of int
                # is one slice in off plane direction (int)
                required_plane_shape.pop(plane_dimension)
            transform = tfm.SizeCorrection(tuple(required_plane_shape), entries=tuple(to_assemble.keys()))
            self.planes[plane_dimension].on_sample_fn = TransformSampleFn(transform)

            self.planes[plane_dimension].add_sample(to_assemble, batch, idx)

        ready = None
        for plane_assembler in self.planes.values():
            if last_batch:
                plane_assembler.end()

            if ready is None:
                ready = set(plane_assembler.subjects_ready)
            else:
                ready.intersection_update(plane_assembler.subjects_ready)
        self._subjects_ready = ready

    def get_assembled_subject(self, subject_index: int):
        """Gets the prediction of a subject.

        Args:
            subject_index (int): The subject's index.

        Returns:
            np.ndarray: The prediction of the subject.
        """
        try:
            self._subjects_ready.remove(subject_index)
        except KeyError:
            # check if subject is assembled but not listed as ready
            # this can happen if only one subject was assembled or last
            if subject_index not in self.planes[0].predictions:
                raise ValueError('Subject with index {} not in assembler'.format(subject_index))

        assembled = {}
        for plane in self.planes.values():
            ret_val = plane.get_assembled_subject(subject_index)
            if not isinstance(ret_val, dict):
                ret_val = {'__prediction': ret_val}
            for key, value in ret_val.items():
                assembled.setdefault(key, []).append(value)

        for key in assembled:
            assembled[key] = self.merge_fn(assembled[key])

        if '__prediction' in assembled:
            return assembled['__prediction']
        return assembled

    @staticmethod
    def _get_plane_dimension(index_expr):
        for i, entry in enumerate(index_expr.expression):
            if isinstance(entry, int):
                return i
            if isinstance(entry, slice) and entry != slice(None):
                return i


class Subject2dAssembler(Assembler):

    def __init__(self, id_entry='ids') -> None:
        super().__init__()
        self._subjects_ready = set()
        self.predictions = {}
        self.id_entry = id_entry

    @property
    def subjects_ready(self):
        return self._subjects_ready

    def add_batch(self, to_assemble, batch: dict, last_batch=False):
        if self.id_entry not in batch:
            raise ValueError('Subject2dAssembler requires "{}" to be in the batch'.format(self.id_entry))

        if not isinstance(to_assemble, dict):
            to_assemble = {'__prediction': to_assemble}

        for idx in range(len(batch[self.id_entry])):
            subject_index = batch[self.id_entry][idx]
            for key in to_assemble:
                self.predictions.setdefault(subject_index, {})[key] = to_assemble[key][idx]
                self._subjects_ready.add(subject_index)

    def get_assembled_subject(self, subject_index):
        try:
            self._subjects_ready.remove(subject_index)
        except KeyError:
            # check if subject is assembled but not listed as ready
            # this can happen if only one subject was assembled or last
            if subject_index not in self.predictions:
                raise ValueError('Subject with index {} not in assembler'.format(subject_index))
        assembled = self.predictions.pop(subject_index)
        if '__prediction' in assembled:
            return assembled['__prediction']
        return assembled




"""The loading module holds classes to load data."""
import abc
import enum
import typing
import os


class FilePathGenerator(metaclass=abc.ABCMeta):
    """Represents an abstract file path generator.

    This class is used in :py:class:`FileSystemDataCrawler` to convert a human readable data identifier to an data file path,
    which allows to load the data.
    """

    @staticmethod
    @abc.abstractmethod
    def get_full_file_path(id_: str, root_dir: str, file_key, file_extension: str) -> str:
        """Gets the full file path for a data file.

        Args:
            id_ (str): The data's identification.
            root_dir (str): The data file's root directory.
            file_key (object): A human readable identifier used to identify the data file.
            file_extension (str): The data's file extension.

        Returns:
            str: The data's full file path.
        """
        raise NotImplementedError()


class DirectoryFilter(metaclass=abc.ABCMeta):
    """Represents an abstract directory filter.

    This class is used in  :py:class:`FileSystemDataCrawler` to filter a list of directories.
    """

    @staticmethod
    @abc.abstractmethod
    def filter_directories(dirs: typing.List[str]) -> typing.List[str]:
        """Filters a list of directories.

        Args:
            dirs (List[str]): A list of directories.

        Returns:
            List[str]: The filtered list of directories.
        """
        raise NotImplementedError()


class FileSystemDataCrawler:
    """Represents a file system data crawler.

    Examples:
        Suppose we have the following directory structure::

            /path/to/root_dir
                ./Patient1
                    ./Image.mha
                    ./GroundTruth.mha
                    ./some_text_file.txt
                ./Patient2
                    ./Image.mha
                    ./GroundTruth.mha
                    ./GroundTruthRater2.mha
                ./Atlas
                    ./Atlas.mha

        We can use the following code to load the images `Image.mha` and `GroundTruth.mha`
        in the directories `Patient1` and `Patient2`:

        >>> class MyImgType(enum.Enum):
        >>>     T1 = 1
        >>>     GroundTruth = 2
        >>>
        >>> class MyFilePathGenerator(FilePathGenerator):
        >>>     @staticmethod
        >>>     def get_full_file_path(id_: str, root_dir: str, file_key, file_extension: str) -> str:
        >>>         if file_key == MyImgType.T1:
        >>>             file_name = 'Image'
        >>>         elif file_key == MyImgType.GroundTruth:
        >>>             file_name = 'GroundTruth'
        >>>         else:
        >>>             raise ValueError('Unknown key')
        >>>
        >>>         return os.path.join(root_dir, file_name + file_extension)
        >>>
        >>> class MyDirFilter(DirectoryFilter):
        >>>     @staticmethod
        >>>     def filter_directories(dirs: typing.List[str]) -> typing.List[str]:
        >>>         return sorted([dir_ for dir_ in dirs if dir_.lower().__contains__('patient')])
        >>>
        >>> crawler = FileSystemDataCrawler('/path/to/root_dir',
        >>>                                 [MyImgType.T1, MyImgType.GroundTruth],
        >>>                                 MyFilePathGenerator(),
        >>>                                 MyDirFilter(),
        >>>                                 '.mha')
        >>> for id_, path in crawler.data.items():
        >>>     print(id_, path)
        Patient1 {'Patient1': '/path/to/root_dir/Patient1', <MyImgType.T1: 1>: '/path/to/root_dir/Patient1/Image.mha', <MyImgType.GroundTruth: 2>: '/path/to/root_dir/Patient1/GroundTruth.mha'}
        Patient2 {'Patient2': '/path/to/root_dir/Patient2', <MyImgType.T1: 1>: '/path/to/root_dir/Patient2/Image.mha', <MyImgType.GroundTruth: 2>: '/path/to/root_dir/Patient2/GroundTruth.mha'}
    """
    
    def __init__(self,
                 root_dir: str,
                 file_keys: list,
                 file_path_generator: FilePathGenerator,
                 dir_filter: DirectoryFilter=None,
                 file_extension: str='.nii.gz'):
        """Initializes a new instance of the FileSystemDataCrawler class.

        Args:
            root_dir (str): The path to the root directory, which contains subdirectories with the data.
            file_keys (list): A list of objects, which represent human readable data identifiers
                (one identifier for each data file to crawl).
            file_path_generator (FilePathGenerator): A file path generator, which converts a human readable
                data identifier to an data file path.
            dir_filter (DirectoryFilter): A directory filter, which filters a list of directories.
            file_extension (str): The data file extension (with or without dot).
        """
        super().__init__()

        self.root_dir = root_dir
        self.dir_filter = dir_filter
        self.file_keys = file_keys
        self.file_path_generator = file_path_generator
        self.file_extension = file_extension if file_extension.startswith('.') else '.' + file_extension

        # dict with key=id (i.e, directory name), value=path to data directory
        self.data = {}  # dict with key=id (i.e, directory name), value=dict with key=file_keys and value=path to file

        data_dir = self._crawl_directories()
        self._crawl_data(data_dir)

    def _crawl_data(self, data_dir: dict):
        """Crawls the data inside a directory."""

        for id_, path in data_dir.items():
            data_dict = {id_: path}  # init dict with id_ pointing to path
            for item in self.file_keys:
                file_path = self.file_path_generator.get_full_file_path(id_, path, item, self.file_extension)
                data_dict[item] = file_path

            self.data[id_] = data_dict

    def _crawl_directories(self):
        """Crawls the directories, which contain data.

        Returns:
            dict: A dictionary where the keys are the directory names and the values the full path to the directory.
        """

        if not os.path.isdir(self.root_dir):
            raise ValueError('root_dir {} does not exist'.format(self.root_dir))

        # search the root directory for data directories
        data_dirs = next(os.walk(self.root_dir))[1]

        if self.dir_filter:
            # filter the data directories
            data_dirs = self.dir_filter.filter_directories(data_dirs)

        return {
            data_dir: os.path.join(self.root_dir, data_dir)
            for data_dir in data_dirs
            if any(file.endswith(self.file_extension) for file  # check if directory contains data files
                   in os.listdir(os.path.join(self.root_dir, data_dir)))
        }



import typing as t


# todo: could maybe be replaced or wrapper with numpy.s_
class IndexExpression:

    def __init__(self, indexing: t.Union[int, tuple, t.List[int], t.List[tuple], t.List[list]]=None,
                 axis: t.Union[int, tuple]=None) -> None:
        self.expression = None
        self.set_indexing(indexing, axis)

    def set_indexing(self, indexing: t.Union[int, tuple, slice, t.List[int], t.List[tuple], t.List[list]],
                     axis: t.Union[int, tuple]=None):
        if indexing is None:
            self.expression = slice(None)
            return

        if isinstance(indexing, int) or isinstance(indexing, tuple):
            indexing = [indexing]

        if axis is None:
            axis = tuple(range(len(indexing)))
        if isinstance(axis, int):
            axis = (axis,)

        expr = [slice(None) for _ in range(max(axis) + 1)]
        for a, index in zip(axis, indexing):
            if isinstance(index, int):
                expr[a] = index
            elif isinstance(index, (tuple, list)):
                start, stop = index
                expr[a] = slice(start, stop)
            elif isinstance(index, slice):
                expr[a] = index
            else:
                raise ValueError('Unknown type "{}" of index'.format(type(index)))

        # needs to be tuple otherwise exception from h5py while slicing
        self.expression = tuple(expr)

    def get_indexing(self):
        indexing = []
        # todo(alainjungo): handle case when self.expression is of type slice
        for index in self.expression:
            if index is None:
                indexing.append(None)
            elif isinstance(index, slice):
                indexing.append((index.start, index.stop))
            elif isinstance(index, int):
                indexing.append(index)
            else:
                raise ValueError("only 'int', 'slice', and 'None' types possible in expression")
        return indexing



"""Enables the plotting of 3-dimensional visualizations.

Use `conda install -c clinicalgraphics vtk` to install the required Visualization Toolkit (VTK) with anaconda.

Examples:
    
    The first example computes the surface-to-surface distance of a unit sphere and unit cube.

    >>> # create sphere and cube
    >>> sphere = vtk.vtkSphereSource()
    >>> sphere.SetPhiResolution(100)
    >>> sphere.SetThetaResolution(100)
    >>> cube = vtk.vtkCubeSource()
    >>> 
    >>> distance_filter = vtk.vtkDistancePolyDataFilter()
    >>> distance_filter.SetInputConnection(0, sphere.GetOutputPort())
    >>> distance_filter.SetInputConnection(1, cube.GetOutputPort())
    >>> distance_filter.Update()
    >>> distance = distance_filter.GetOutput()
    
    The second example computes the surface-to-surface distance of a prediction to a ground truth image.

    >>> image1, transform1 = read_and_transform_image('PREDICTION.mha')
    >>> image2, transform2 = read_and_transform_image('GROUND-TRUTH.mha')
    >>> isocontour1 = compute_isocontour(image1)
    >>> isocontour2 = compute_isocontour(image2)
    >>> model = compute_surface_to_surface_distance(isocontour1, isocontour2)

    
"""
import pathlib

import SimpleITK as sitk
try:
    import vtk
except ImportError:
    raise ImportError('Module "vtk" not found. Install "vtk==8.1.2" to use the surfacedistance module.')


def compute_isocontour(image_label: vtk.vtkImageData, label: int=1) -> vtk.vtkPolyData:
    """Calculates the isocontour from a label image.

    Args:
        image_label (vtk.vtkImageData): The label image.
        label (int): The label.

    Returns:
        (vtk.vtkPolyData): The extracted isocontour.
    """
    marchingCubes = vtk.vtkMarchingCubes()
    marchingCubes.SetInputData(image_label)
    marchingCubes.SetValue(0, label)
    marchingCubes.Update()

    return marchingCubes.GetOutput()


def compute_surface_to_surface_distance(isocontour1: vtk.vtkPolyData, isocontour2: vtk.vtkPolyData) -> vtk.vtkPolyData:
    """Calculates the surface-to-surface distance between two isocontours.

    Args:
        isocontour1 (vtk.vtkPolyData): The first isocontour.
        isocontour2 (vtk.vtkPolyData): The second isocontour.

    Returns:
        (vtk.vtkPolyData)
    """
    cleaner1 = vtk.vtkCleanPolyData()
    cleaner1.SetInputData(isocontour1)

    cleaner2 = vtk.vtkCleanPolyData()
    cleaner2.SetInputData(isocontour2)

    distance_filter = vtk.vtkDistancePolyDataFilter()
    distance_filter.SetInputConnection(0, cleaner1.GetOutputPort())
    distance_filter.SetInputConnection(1, cleaner2.GetOutputPort())
    distance_filter.Update()

    return distance_filter.GetOutput()


def read_and_transform_image(file_path: str) -> (vtk.vtkImageData, vtk.vtkTransform):
    """Reads an image as VTK image.

    VTK images have no direction matrix, therefore we apply a transformation to the image after loading to have in the
    correct space. Note that only meta and NIfTI images are supported.

    Args:
        file_path (str): The file path.

    Returns:
        (vtk.vtkImageData): The image.

    Raises:
        ValueError:
    """
    # check file extension to use correct reader
    extension = ''.join(pathlib.Path(file_path).suffixes)

    if extension == '.mha' or extension == '.mdh':
        reader = vtk.vtkMetaImageReader()
    elif extension == '.nii' or extension == '.nii.gz':
        reader = vtk.vtkNIFTIImageReader()
    else:
        raise ValueError('Unknown image format {}'.format(extension))

    reader.SetFileName(file_path)
    reader.Update()
    image = reader.GetOutput()

    # VTK images have no direction matrix, therefore load the same image with itk and apply a transformation
    image_sitk = sitk.ReadImage(file_path)
    direction = image_sitk.GetDirection()
    origin = image_sitk.GetOrigin()

    matrix = vtk.vtkMatrix4x4()
    matrix.Identity()
    matrix.SetElement(0, 0, direction[0])
    matrix.SetElement(0, 1, direction[3])
    matrix.SetElement(0, 2, direction[6])
    matrix.SetElement(1, 0, direction[1])
    matrix.SetElement(1, 1, direction[4])
    matrix.SetElement(1, 2, direction[7])
    matrix.SetElement(2, 0, direction[2])
    matrix.SetElement(2, 1, direction[5])
    matrix.SetElement(2, 2, direction[8])

    transform = vtk.vtkTransform()
    transform.Identity()
    # transform.Update()
    # transform.PostMultiply()
    # transform.Translate(-origin[0], -origin[1], -origin[2])
    transform.Translate(-origin[0], -origin[1], -origin[2])
    transform.Concatenate(matrix)

    return image, transform



import abc
import typing as t

import numpy as np

from pymia.data import subjectfile as subj
import pymia.data.transformation as tfm
import pymia.data.conversion as conv
from . import callback as cb
from . import fileloader as load


class Traverser(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def traverse(self, files: list, loader: load.Load, callbacks: t.List[cb.Callback]=None,
                 transform: tfm.Transform=None):
        pass


def default_concat(data: t.List[np.ndarray]) -> np.ndarray:
    return np.stack(data, axis=-1)


class SubjectFileTraverser(Traverser):

    def __init__(self, categories: t.Union[str, t.Tuple[str, ...]]=None):
        """Initializes a new instance of the SubjectFileTraverser class.

        Args:
            categories (str or tuple of str): The categories to traverse. If None, then all categories of a SubjectFile
                will be traversed.
        """
        if isinstance(categories, str):
            categories = (categories, )
        self.categories = categories

    def traverse(self, subject_files: t.List[subj.SubjectFile], load=load.LoadDefault(), callback: cb.Callback=None,
                 transform: tfm.Transform=None, concat_fn=default_concat):
        if len(subject_files) == 0:
            raise ValueError('No files')
        if not isinstance(subject_files[0], subj.SubjectFile):
            raise ValueError('files must be of type {}'.format(subj.SubjectFile.__class__.__name__))
        if callback is None:
            raise ValueError('callback can not be None')

        if self.categories is None:
            self.categories = subject_files[0].categories

        callback_params = {'subject_files': subject_files}
        for category in self.categories:
            callback_params.setdefault('categories', []).append(category)
            callback_params['{}_names'.format(category)] = self._get_names(subject_files, category)
        callback.on_start(callback_params)

        # looping over the subject files and calling callbacks
        for subject_index, subject_file in enumerate(subject_files):
            transform_params = {'subject_index': subject_index}
            for category in self.categories:

                category_list = []
                category_property = None  # type: conv.ImageProperties
                for id_, file_path in subject_file.categories[category].entries.items():
                    np_data, data_property = load(file_path, id_, category, subject_file.subject)
                    category_list.append(np_data)
                    if category_property is None:  # only required once
                        category_property = data_property

                category_data = concat_fn(category_list)
                transform_params[category] = category_data
                transform_params['{}_properties'.format(category)] = category_property

            if transform:
                transform_params = transform(transform_params)

            callback.on_subject({**transform_params, **callback_params})

        callback.on_end(callback_params)

    @staticmethod
    def _get_names(subject_files: t.List[subj.SubjectFile], category: str) -> list:
        names = subject_files[0].categories[category].entries.keys()
        if not all(s.categories[category].entries.keys() == names for s in subject_files):
            raise ValueError('Inconsistent {} identifiers in the subject list'.format(category))
        return list(names)



"""This module holds classes related to image conversion.

The main purpose of this module is the conversion between SimpleITK images and numpy arrays.
"""
import typing

import SimpleITK as sitk
import numpy as np


class ImageProperties:
    """Represents ITK image properties.

    Holds common ITK image meta-data such as the size, origin, spacing, and direction.

    See Also:
        SimpleITK provides `itk::simple::Image::CopyInformation`_ to copy image information.

    .. _itk::simple::Image::CopyInformation:
        https://itk.org/SimpleITKDoxygen/html/classitk_1_1simple_1_1Image.html#afa8a4757400c414e809d1767ee616bd0
    """

    def __init__(self, image: sitk.Image):
        """Initializes a new instance of the ImageProperties class.

        Args:
            image (sitk.Image): The image whose properties to hold.
        """
        self.size = image.GetSize()
        self.origin = image.GetOrigin()
        self.spacing = image.GetSpacing()
        self.direction = image.GetDirection()
        self.dimensions = image.GetDimension()
        self.number_of_components_per_pixel = image.GetNumberOfComponentsPerPixel()
        self.pixel_id = image.GetPixelID()

    def is_two_dimensional(self) -> bool:
        """Determines whether the image is two-dimensional.

        Returns:
            bool: True if the image is two-dimensional; otherwise, False.
        """
        return self.dimensions == 2

    def is_three_dimensional(self) -> bool:
        """Determines whether the image is three-dimensional.

        Returns:
            bool: True if the image is three-dimensional; otherwise, False.
        """
        return self.dimensions == 3

    def is_vector_image(self) -> bool:
        """Determines whether the image is a vector image.

        Returns:
            bool: True for vector images; False for scalar images.
        """
        return self.number_of_components_per_pixel > 1

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'ImageProperties:\n' \
               ' size:                           {self.size}\n' \
               ' origin:                         {self.origin}\n' \
               ' spacing:                        {self.spacing}\n' \
               ' direction:                      {self.direction}\n' \
               ' dimensions:                     {self.dimensions}\n' \
               ' number_of_components_per_pixel: {self.number_of_components_per_pixel}\n' \
               ' pixel_id:                       {self.pixel_id}\n' \
            .format(self=self)

    def __eq__(self, other):
        """Determines the equality of two ImageProperties classes.

        Notes
            The equality does not include the number_of_components_per_pixel and pixel_id.

        Args:
            other (object): An ImageProperties instance or any other object.

        Returns:
            bool: True if the ImageProperties are equal; otherwise, False.
        """
        if isinstance(other, self.__class__):
            return self.size == other.size and \
                   self.origin == other.origin and \
                   self.spacing == other.spacing and \
                   self.direction == other.direction and \
                   self.dimensions == other.dimensions
        return NotImplemented

    def __ne__(self, other):
        """Determines the non-equality of two ImageProperties classes.

        Notes
            The non-equality does not include the number_of_components_per_pixel and pixel_id.

        Args:
            other (object): An ImageProperties instance or any other object.

        Returns:
            bool: True if the ImageProperties are non-equal; otherwise, False.
        """
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        """Gets the hash.

        Returns:
            int: The hash of the object.
        """
        return hash(tuple(sorted(self.__dict__.items())))


class NumpySimpleITKImageBridge:
    """A numpy to SimpleITK bridge, which provides static methods to convert between numpy array and SimpleITK image."""

    @staticmethod
    def convert(array: np.ndarray, properties: ImageProperties) -> sitk.Image:
        """Converts a numpy array to a SimpleITK image.

        Args:
            array (np.ndarray): The image as numpy array. The shape can be either:

                - shape=(n,), where n = total number of voxels
                - shape=(n,v), where n = total number of voxels and v = number of components per pixel (vector image)
                - shape=(<reversed image size>), what you get from sitk.GetArrayFromImage()
                - shape=(<reversed image size>,v), what you get from sitk.GetArrayFromImage()
                    and v = number of components per pixel (vector image)

            properties (ImageProperties): The image properties.

        Returns:
            sitk.Image: The SimpleITK image.
        """

        is_vector = False
        if not array.shape == properties.size[::-1]:
            # we need to reshape the array

            if array.ndim == 1:
                array = array.reshape(properties.size[::-1])
            elif array.ndim == 2:
                is_vector = True
                array = array.reshape((properties.size[::-1] + (array.shape[1],)))
            elif array.ndim == len(properties.size) + 1:
                is_vector = True
                # no need to reshape
            else:
                raise ValueError('array shape {} not supported'.format(array.shape))

        image = sitk.GetImageFromArray(array, is_vector)
        image.SetOrigin(properties.origin)
        image.SetSpacing(properties.spacing)
        image.SetDirection(properties.direction)

        return image


class SimpleITKNumpyImageBridge:
    """A SimpleITK to numpy bridge.

    Converts SimpleITK images to numpy arrays. Use the ``NumpySimpleITKImageBridge`` to convert back.
    """

    @staticmethod
    def convert(image: sitk.Image) -> typing.Tuple[np.ndarray, ImageProperties]:
        """Converts an image to a numpy array and an ImageProperties class.

        Args:
            image (SimpleITK.Image): The image.

        Returns:
            A Tuple[np.ndarray, ImageProperties]: The image as numpy array and the image properties.

        Raises:
            ValueError: If `image` is `None`.
        """

        if image is None:
            raise ValueError('Parameter image can not be None')

        return sitk.GetArrayFromImage(image), ImageProperties(image)



"""The post-processing module contains classes for image filtering mostly applied after a classification.

Image post-processing aims to alter images such that they depict a desired representation.
"""
import SimpleITK as sitk

import pymia.filtering.filter as pymia_fltr


class BinaryThreshold(pymia_fltr.IFilter):
    """Represents a binary threshold image filter."""

    def __init__(self, threshold: float):
        """Initializes a new instance of the BinaryThreshold class.

        Args:
            threshold: The threshold value.
        """
        super().__init__()
        self.threshold = threshold
        self.filter = sitk.BinaryThresholdImageFilter()
        self.filter.SetInsideValue(0)
        self.filter.SetOutsideValue(1)
        self.filter.SetUpperThreshold(self.threshold)

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams = None) -> sitk.Image:
        """Executes the binary threshold filter on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The filtered image.
        """
        return self.filter.Execute(image)


class LargestNConnectedComponents(pymia_fltr.IFilter):
    """Represents a largest N connected components filter.

    Extracts the largest N connected components from a label image.
    By default the N components will all have the value 1 in the output image.
    Use the `consecutive_component_labels` option such that the largest has value 1,
    the second largest has value 2, etc. Background is always assumed to be 0.
    """

    def __init__(self, number_of_components: int = 1, consecutive_component_labels: bool = False):
        """Initializes a new instance of the LargestNComponents class.

        Args:
            number_of_components (int): The number of largest components to extract.
            consecutive_component_labels (bool): The largest component has value 1, the second largest has value 2, ect.
                if set to True; otherwise, all components will have value 1.
        """
        super().__init__()

        if not number_of_components >= 1:
            raise ValueError("number_of_components must be larger or equal to 1")

        self.number_of_components = number_of_components
        self.consecutive_component_labels = consecutive_component_labels

    def execute(self, image: sitk.Image, params: pymia_fltr.IFilterParams = None) -> sitk.Image:
        """Executes the largest N connected components filter on an image.

        Args:
            image (sitk.Image): The image.
            params (IFilterParams): The parameters (unused).

        Returns:
            sitk.Image: The filtered image.
        """

        image = sitk.ConnectedComponent(image)
        image = sitk.RelabelComponent(image)

        if self.consecutive_component_labels:
            return sitk.Threshold(image, lower=1, upper=self.number_of_components, outsideValue=0)
        else:
            return sitk.BinaryThreshold(image, lowerThreshold=1, upperThreshold=self.number_of_components,
                                        insideValue=1, outsideValue=0)

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """

        return 'LargestNConnectedComponents:\n' \
               ' number_of_components:         {self.number_of_components}\n' \
               ' consecutive_component_labels: {self.consecutive_component_labels}\n' \
            .format(self=self)



import torch.utils.data.dataset as data

import pymia.data.transformation as tfm
import pymia.data.indexexpression as expr
from . import reader as rd
from . import indexing as idx
from . import extractor as extr


class ParameterizableDataset(data.Dataset):

    def __init__(self, dataset_path: str, indexing_strategy: idx.IndexingStrategy=None, extractor: extr.Extractor=None,
                 transform: tfm.Transform=None, subject_subset: list=None, init_reader_once=True) -> None:
        self.dataset_path = dataset_path
        self.indexing_strategy = None
        self.extractor = extractor
        self.transform = transform
        self.subject_subset = subject_subset
        self.init_reader_once = init_reader_once
        self.indices = []
        self.reader = None

        # init indices
        if indexing_strategy is None:
            indexing_strategy = idx.EmptyIndexing()
        self.set_indexing_strategy(indexing_strategy, subject_subset)

    def close_reader(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None

    def set_extractor(self, extractor: extr.Extractor):
        self.extractor = extractor

    def set_indexing_strategy(self, indexing_strategy: idx.IndexingStrategy, subject_subset: list = None):
        self.indices.clear()
        self.indexing_strategy = indexing_strategy
        with rd.get_reader(self.dataset_path) as reader:
            all_subjects = reader.get_subjects()
            last_shape = None  # remember shape to optimize initialization
            for i, subject in enumerate(reader.get_subject_entries()):
                if subject_subset is None or all_subjects[i] in subject_subset:
                    current_shape = reader.get_shape(subject)
                    if not last_shape == current_shape:
                        subject_indices = self.indexing_strategy(current_shape)
                        last_shape = current_shape

                    subject_and_indices = zip(len(subject_indices) * [i], subject_indices)
                    self.indices.extend(subject_and_indices)

    def set_transform(self, transform: tfm.Transform):
        self.transform = transform

    def get_subjects(self):
        with rd.get_reader(self.dataset_path) as reader:
            return reader.get_subjects()

    def direct_extract(self, extractor: extr.Extractor, subject_index: int, index_expr: expr.IndexExpression=None,
                       transform: tfm.Transform=None):
        if index_expr is None:
            index_expr = expr.IndexExpression()

        params = {'subject_index': subject_index, 'index_expr': index_expr}
        extracted = {}

        if not self.init_reader_once:
            with rd.get_reader(self.dataset_path) as reader:
                extractor.extract(reader, params, extracted)
        else:
            if self.reader is None:
                self.reader = rd.get_reader(self.dataset_path, direct_open=True)
            extractor.extract(self.reader, params, extracted)

        if transform:
            extracted = transform(extracted)

        return extracted

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, item):
        subject_index, index_expr = self.indices[item]
        return self.direct_extract(self.extractor, subject_index, index_expr, self.transform)

    def __del__(self):
        self.close_reader()



"""Enables the plotting of image histograms."""
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import typing as t
import SimpleITK as sitk


def plot_histogram(path: str, image: t.Union[sitk.Image, np.ndarray], no_bins: int=255, use_percentages: bool=False,
                   title: str='', xlabel: str='', ylabel: str='', slice_no: int=-1) -> None:
    """Plots a histogram of an image.

    Plots either the histogram of a slice of the image or of the whole image.

    Args:
        path (str): The file path.
        image (SimpleITK.Image or numpy array): The image.
        no_bins (int): The number of histogram bins.
        use_percentages (bool): Use percentages on the y axis if True; otherwise, use the number of values in the bins.
        title (str): The histogram's title.
        xlabel (str): The histogram's x-axis label.
        ylabel (str): The histogram's y-axis label.
        slice_no (int): The slice number or -1 to take the whole image (only supported for SimpleITK images).
    """

    if isinstance(image, sitk.Image):
        if slice_no > -1:
            data = sitk.GetArrayFromImage(image[:, :, slice_no])
        else:
            data = sitk.GetArrayFromImage(image)
    else:
        data = image

    data = data.flatten()

    plt.hist(data, bins=no_bins, density=use_percentages)
    if title: plt.title(title)
    if xlabel: plt.xlabel(xlabel)
    if ylabel: plt.ylabel(ylabel)

    if use_percentages:
        ax = plt.gca()
        ax.yaxis.set_major_formatter(ticker.PercentFormatter(xmax=1))

    plt.savefig(path)
    plt.close()


def plot_histogram_overlay(path: str, image1: sitk.Image, image2: sitk.Image,
                           legend1: str = '', legend2: str = '',
                           no_bins: int = 255, slice_no: int = -1,
                           title: str = '', xlabel: str = '', ylabel: str = '',
                           xaxis_log: bool = False, yaxis_log: bool = False) -> None:
    """Plots a histogram of an image.

    Plots either the histogram of a slice of the image or of the whole image.

    Args:
        path (str): The file path.
        image1 (sitk.Image): The first image.
        image2 (sitk.Image): The second image.
        legend1 (str): The legend for the first image.
        legend2 (str): The legend for the second image.2
        no_bins (int): The number of histogram bins.
        slice_no (int): The slice number or -1 to take the whole image.
        title (str): The histogram's title.
        xlabel (str): The histogram's x-axis label.
        ylabel (str): The histogram's y-axis label.
        xaxis_log (bool): Logarithmic scale for x-axis.
        yaxis_log (bool): Logarithmic scale for y-axis.
    """
    if slice_no > -1:
        data1 = sitk.GetArrayFromImage(image1[:, :, slice_no])
        data2 = sitk.GetArrayFromImage(image2[:, :, slice_no])
    else:
        data1 = sitk.GetArrayFromImage(image1)
        data2 = sitk.GetArrayFromImage(image2)

    data1 = data1.flatten()
    data2 = data2.flatten()

    plt.hist(data1, bins=no_bins, alpha=0.5, label=legend1)
    plt.hist(data2, bins=no_bins, alpha=0.5, label=legend2)
    if legend1 or legend2: plt.legend()
    if xaxis_log: plt.xscale('log')
    if yaxis_log: plt.yscale('log')
    if title: plt.title(title)
    if xlabel: plt.xlabel(xlabel)
    if ylabel: plt.ylabel(ylabel)
    plt.savefig(path)
    plt.close()



import abc
import typing

import numpy as np
import torch


ENTRY_NOT_EXTRACTED_ERR_MSG = 'Transform can not be applied because entry "{}" was not extracted'

raise_error_if_entry_not_extracted = True


# follows the principle of torchvision transform
class Transform(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __call__(self, sample: dict) -> dict:
        pass


class ComposeTransform(Transform):

    def __init__(self, transforms: typing.Iterable[Transform]) -> None:
        self.transforms = transforms

    def __call__(self, sample: dict) -> dict:
        for t in self.transforms:
            sample = t(sample)
        return sample


class IntensityRescale(Transform):

    def __init__(self, lower, upper, loop_axis=None, entries=('images',)) -> None:
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.loop_axis = loop_axis
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            if self.loop_axis is None:
                np_entry = self._normalize(np_entry, self.lower, self.upper)
            else:
                slicing = [slice(None) for _ in range(np_entry.ndim)]
                for i in range(np_entry.shape[self.loop_axis]):
                    slicing[self.loop_axis] = i
                    np_entry[tuple(slicing)] = self._normalize(np_entry[tuple(slicing)], self.lower, self.upper)

            sample[entry] = np_entry
        return sample

    @staticmethod
    def _normalize(arr: np.ndarray, lower, upper):
        dtype = arr.dtype
        min_, max_ = arr.min(), arr.max()
        if min_ == max_:
            raise ValueError('cannot normalize when min == max')
        arr = (arr - min_) / (max_ - min_) * (upper - lower) + lower
        return arr.astype(dtype)


class IntensityNormalization(Transform):

    def __init__(self, loop_axis=None, entries=('images',)) -> None:
        super().__init__()
        self.loop_axis = loop_axis
        self.entries = entries
        self.normalize_fn = self._normalize

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            if not np.issubdtype(np_entry.dtype, np.floating):
                raise ValueError('Array must be floating type')

            if self.loop_axis is None:
                np_entry = self.normalize_fn(np_entry)
            else:
                slicing = [slice(None) for _ in range(np_entry.ndim)]
                for i in range(np_entry.shape[self.loop_axis]):
                    slicing[self.loop_axis] = i
                    np_entry[tuple(slicing)] = self.normalize_fn(np_entry[tuple(slicing)])
            sample[entry] = np_entry
        return sample

    @staticmethod
    def _normalize(arr: np.ndarray):
        return (arr - arr.mean()) / arr.std()


class LambdaTransform(Transform):

    def __init__(self, lambda_fn, loop_axis=None, entries=('images',)) -> None:
        super().__init__()
        self.lambda_fn = lambda_fn
        self.loop_axis = loop_axis
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            if self.loop_axis is None:
                np_entry = self.lambda_fn(sample[entry])
            else:
                np_entry = check_and_return(sample[entry], np.ndarray)
                slicing = [slice(None) for _ in range(np_entry.ndim)]
                for i in range(np_entry.shape[self.loop_axis]):
                    slicing[self.loop_axis] = i
                    np_entry[tuple(slicing)] = self.lambda_fn(np_entry[tuple(slicing)])
            sample[entry] = np_entry
        return sample


class ClipPercentile(Transform):

    def __init__(self, upper_percentile: float, lower_percentile: float=None,
                 loop_axis=None, entries=('images',)) -> None:
        super().__init__()
        self.upper_percentile = upper_percentile
        if lower_percentile is None:
            lower_percentile = 100 - upper_percentile
        self.lower_percentile = lower_percentile
        self.loop_axis = loop_axis
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)

            if self.loop_axis is None:
                np_entry = self._clip(np_entry)
            else:
                slicing = [slice(None) for _ in range(np_entry.ndim)]
                for i in range(np_entry.shape[self.loop_axis]):
                    slicing[self.loop_axis] = i
                    np_entry[tuple(slicing)] = self._clip(np_entry[tuple(slicing)])

            sample[entry] = np_entry
        return sample

    def _clip(self, arr: np.ndarray):
        upper_max = np.percentile(arr, self.upper_percentile)
        arr[arr > upper_max] = upper_max
        lower_max = np.percentile(arr, self.lower_percentile)
        arr[arr < lower_max] = lower_max
        return arr


class Relabel(Transform):

    def __init__(self, label_changes: typing.Dict[int, int], entries=('labels',)) -> None:
        super().__init__()
        self.label_changes = label_changes
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            for new_label, old_label in self.label_changes.items():
                np_entry[np_entry == old_label] = new_label
            sample[entry] = np_entry
        return sample


class Reshape(Transform):

    def __init__(self, shapes: dict) -> None:
        """Initializes a new instance of the Reshape class.

        Args:
            shapes (dict): A dict with keys being the entries and the values the new shapes of the entries.
                E.g. shapes = {'images': (-1, 4), 'labels' : (-1, 1)}
        """
        super().__init__()
        self.shapes = shapes

    def __call__(self, sample: dict) -> dict:
        for entry in self.shapes:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            sample[entry] = np.reshape(np_entry, self.shapes[entry])
        return sample


class ToTorchTensor(Transform):

    def __init__(self, entries=('images', 'labels')) -> None:
        super().__init__()
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            sample[entry] = torch.from_numpy(np_entry)
        return sample


class Permute(Transform):

    def __init__(self, permutation: tuple, entries=('images', 'labels')) -> None:
        super().__init__()
        self.permutation = permutation
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            sample[entry] = np.transpose(np_entry, self.permutation)
        return sample


class Squeeze(Transform):

    def __init__(self, entries=('images', 'labels'), squeeze_axis=None) -> None:
        super().__init__()
        self.entries = entries
        self.squeeze_axis = squeeze_axis

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            sample[entry] = np_entry.squeeze(self.squeeze_axis)
        return sample


class UnSqueeze(Transform):

    def __init__(self, axis=-1, entries=('images', 'labels')) -> None:
        super().__init__()
        self.entries = entries
        self.axis = axis

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            sample[entry] = np.expand_dims(np_entry, self.axis)
        return sample


class SizeCorrection(Transform):
    """Size correction transformation.

    Corrects the size, i.e. shape, of an array to a given reference shape.
    """

    def __init__(self, shape: typing.Tuple[typing.Union[None, int], ...], pad_value: int=0,
                 entries=('images', 'labels')) -> None:
        """Initializes a new instance of the SizeCorrection class.

        Args:
            shape (tuple of ints): The reference shape in NumPy format, i.e. z-, y-, x-order. To not correct an axis
                dimension, set the axis value to None.
            pad_value (int): The value to set the padded values of the array.
            entries ():
        """
        super().__init__()
        self.entries = entries
        self.shape = shape
        self.pad_value = pad_value

    def __call__(self, sample: dict) -> dict:
        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)
            if not len(self.shape) <= np_entry.ndim:
                raise ValueError('Shape dimension needs to be less or equal to {}'.format(np_entry.ndim))

            for idx, size in enumerate(self.shape):
                if size is not None and size < np_entry.shape[idx]:
                    # crop current dimension
                    before = (np_entry.shape[idx] - size) // 2
                    after = np_entry.shape[idx] - (np_entry.shape[idx] - size) // 2 - ((np_entry.shape[idx] - size) % 2)
                    slicing = [slice(None)] * np_entry.ndim
                    slicing[idx] = slice(before, after)
                    np_entry = np_entry[tuple(slicing)]
                elif size is not None and size > np_entry.shape[idx]:
                    # pad current dimension
                    before = (size - np_entry.shape[idx]) // 2
                    after = (size - np_entry.shape[idx]) // 2 + ((size - np_entry.shape[idx]) % 2)
                    pad_width = [(0, 0)] * np_entry.ndim
                    pad_width[idx] = (before, after)
                    np_entry = np.pad(np_entry, pad_width, mode='constant', constant_values=self.pad_value)

            sample[entry] = np_entry

        return sample


class Mask(Transform):

    def __init__(self, mask_key: str, mask_value: int=0, masking_value: float=0,
                 loop_axis=None, entries=('images', 'labels')) -> None:
        super().__init__()
        self.mask_key = mask_key
        self.mask_value = mask_value
        self.masking_value = masking_value
        self.loop_axis = loop_axis
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        np_mask = check_and_return(sample[self.mask_key], np.ndarray)

        for entry in self.entries:
            if entry not in sample:
                if raise_error_if_entry_not_extracted:
                    raise ValueError(ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))
                continue

            np_entry = check_and_return(sample[entry], np.ndarray)

            if np_mask.shape == np_entry.shape:
                np_entry[np_mask == self.mask_value] = self.masking_value
            else:
                mask_for_np_entry = np.repeat(np.expand_dims(np_mask, self.loop_axis),
                                              np_entry.shape[self.loop_axis], axis=self.loop_axis)
                np_entry[mask_for_np_entry == self.mask_value] = self.masking_value

            sample[entry] = np_entry
        return sample


def check_and_return(obj, type_):
    if not isinstance(obj, type_):
        raise ValueError("entry must be '{}'".format(type_.__name__))
    return obj



import abc
import typing as t

import numpy as np
import torch.utils.data as data
import torch.utils.data.sampler as smplr


class SelectionStrategy(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __call__(self, sample) -> bool:
        pass

    def __repr__(self) -> str:
        return self.__class__.__name__


class NonConstantSelection(SelectionStrategy):

    def __init__(self, loop_axis=None) -> None:
        super().__init__()
        self.loop_axis = loop_axis

    def __call__(self, sample) -> bool:
        image_data = sample['images']

        if self.loop_axis is None:
            return not self._all_equal(image_data)

        slicing = [slice(None) for _ in range(image_data.ndim)]
        for i in range(image_data.shape[self.loop_axis]):
            slicing[self.loop_axis] = i
            slice_data = image_data[slicing]
            if not self._all_equal(slice_data):
                return True
        return False

    @staticmethod
    def _all_equal(image_data):
        return np.all(image_data == image_data.ravel()[0])

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.loop_axis)


class NonBlackSelection(SelectionStrategy):

    def __init__(self, black_value: float=.0) -> None:
        # todo: add something to do this for one sequence only, too -> loop over one dim
        self.black_value = black_value

    def __call__(self, sample) -> bool:
        return (sample['images'] > self.black_value).any()

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.black_value)


class PercentileSelection(SelectionStrategy):

    def __init__(self, percentile: float) -> None:
        # todo: add something to do this for one sequence only, too -> loop over one dim
        self.percentile = percentile

    def __call__(self, sample) -> bool:
        image_data = sample['images']

        percentile_value = np.percentile(image_data, self.percentile)
        return (image_data >= percentile_value).all()

    def __repr__(self) -> str:
        return '{} ({})'.format(self.__class__.__name__, self.percentile)


class WithForegroundSelection(SelectionStrategy):

    def __call__(self, sample) -> bool:
        return (sample['labels']).any()


class SubjectSelection(SelectionStrategy):
    """Select subjects by their name or index."""

    def __init__(self, subjects) -> None:
        if isinstance(subjects, int):
            subjects = (subjects, )
        if isinstance(subjects, str):
            subjects = (subjects, )
        self.subjects = subjects

    def __call__(self, sample) -> bool:
        return sample['subject'] in self.subjects or sample['subject_index'] in self.subjects

    def __repr__(self) -> str:
        return '{} ({})'.format(self.__class__.__name__, ','.join(self.subjects))


class ComposeSelection(SelectionStrategy):

    def __init__(self, strategies) -> None:
        self.strategies = strategies

    def __call__(self, sample) -> bool:
        return all(strategy(sample) for strategy in self.strategies)

    def __repr__(self) -> str:
        return '|'.join(repr(s) for s in self.strategies)


def select_indices(data_source: data.Dataset, selection_strategy: SelectionStrategy):
    selected_indices = []
    for i, sample in enumerate(data_source):
        if selection_strategy(sample):
            selected_indices.append(i)
    return selected_indices


class SubsetSequentialSampler(smplr.Sampler):
    """Samples elements sequential from a given list of indices, without replacement."""

    def __init__(self, indices):
        super().__init__(None)
        self.indices = indices

    def __iter__(self):
        return (idx for idx in self.indices)

    def __len__(self):
        return len(self.indices)



"""This module holds classes for data augmentation.

The data augmentation bases on the transformation concept (see :class:`pymia.data.transformation.Transform`)
and can easily be incorporated into the data loading process.

Warnings:
The augmentation relies on the random number generator of `numpy`. If you want to obtain reproducible result,
set numpy's seed prior to executing any augmentation:

>>> import numpy as np
>>> your_seed = 0
>>> np.random.seed(your_seed)

See Also:
    https://github.com/MIC-DKFZ/batchgenerators
"""
import typing

import numpy as np

import pymia.data.transformation as tfm
import SimpleITK as sitk


class RandomCrop(tfm.Transform):

    def __init__(self, shape: typing.Union[int, tuple], axis: typing.Union[int, tuple]=None,
                 p: float=1.0, entries=('images', 'labels')):
        """Randomly crops the sample to the specified shape.

        The sample shape must be bigger than the crop shape.

        Notes:
            A probability lower than 1.0 might make not much sense because it results in inconsistent output dimensions.

        Args:
            shape (int, tuple): The shape of the sample after the cropping.
                If axis is not defined, the cropping will be applied from the first dimension onwards of the sample.
                Use None to exclude an axis or define axis to specify the axis/axes to crop.
                E.g.:
                    shape=256 with the default axis parameter results in a shape of 256 x ...
                    shape=(256, 128) with the default axis parameter results in a shape of 256 x 128 x ...
                    shape=(None, 256) with the default axis parameter results in a shape of <as before> x 256 x ...
                    shape=(256, 128) with axis=(1, 0) results in a shape of 128 x 256 x ...
                    shape=(None, 128, 256) with axis=(1, 2, 0) results in a shape of 256 x <as before> x 256 x ...
            axis (int, tuple): Axis or axes to which the shape int or tuple correspond(s) to.
                If defined, must have the same length as shape.
            p (float): The probability of the cropping to be applied.
            entries (tuple): The sample's entries to apply the cropping to.
        """
        super().__init__()
        if isinstance(shape, int):
            shape = (shape, )

        if axis is None:
            axis = tuple(range(len(shape)))
        if isinstance(axis, int):
            axis = (axis, )

        if len(axis) != len(shape):
            raise ValueError('If specified, the axis parameter must be of the same length as the shape')

        # filter out any axis where shape is None
        self.axis = tuple([a for a, s in zip(axis, shape) if s is not None])
        self.shape = tuple([s for s in shape if s is not None])

        self.p = p
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        if self.p < np.random.random():
            return sample

        for entry in self.entries:
            if entry not in sample:
                raise ValueError(tfm.ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))

        anchors = [np.random.randint(0, sample[self.entries[0]].shape[a] - s) for a, s in zip(self.axis, self.shape)]

        for entry in self.entries:
            # todo(fabianbalsiger): replace by slicing (more elegant and faster?)
            for axis, new_axis_size, anchor in zip(self.axis, self.shape, anchors):
                sample[entry] = np.take(sample[entry], range(anchor, anchor + new_axis_size), axis)

        return sample


class RandomElasticDeformation(tfm.Transform):

    def __init__(self, num_control_points: int=4, deformation_sigma=15,
                 interpolators: tuple=(sitk.sitkBSpline, sitk.sitkNearestNeighbor),
                 spatial_rank: int=2, fill_value: float=0.0,
                 p: float=0.5, entries=('images', 'labels')):
        """Randomly transforms the sample elastically.

        Notes:
            The code bases on NiftyNet's RandomElasticDeformationLayer class (version 0.3.0).

        Warnings:
            Always inspect the results of this transform on some samples (especially for 3-D data).

        Args:
            num_control_points (int): The number of control points for the b-spline mesh.
            deformation_sigma: The maximum deformation along the deformation mesh.
            interpolators (tuple): The SimpleITK interpolators to use for each entry in entries.
            spatial_rank (int): The spatial rank (dimension) of the sample.
            fill_value (float): The fill value for the resampling.
            p (float): The probability of the elastic transformation to be applied.
            entries (tuple): The sample's entries to apply the elastic transformation to.
        """
        super().__init__()
        if len(interpolators) != len(entries):
            raise ValueError('interpolators must have the same length as entries')

        self.num_control_points = max(num_control_points, 2)  # need at minimum 2 control points
        self.deformation_sigma = max(deformation_sigma, 1)  # need at minimum 1
        self.spatial_rank = spatial_rank
        self.interpolators = interpolators
        self.fill_value = fill_value

        self.p = p
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        if self.p < np.random.random():
            return sample

        for entry in self.entries:
            if entry not in sample:
                raise ValueError(tfm.ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))

        # initialize a SimpleITK image
        shape = sample[self.entries[0]].shape[:self.spatial_rank]
        img = sitk.GetImageFromArray(np.zeros(shape))  # todo(fabianbalsiger): set spacing etc with ImagePropertiesExtractor?

        # initialize B-spline transformation
        transform_mesh_size = [self.num_control_points] * img.GetDimension()  # todo(fabianbalsiger): allow control points to be defined per image dimension. Allow None to do no deformation in the direction (see comment below)
        bspline_transformation = sitk.BSplineTransformInitializer(img, transform_mesh_size)
        params = bspline_transformation.GetParameters()
        params = np.asarray(params, dtype=np.float)
        params += np.random.randn(params.shape[0]) * self.deformation_sigma

        # remove z deformations! The resolution in z is too bad
        # params[0:int(len(params) / 3)] = 0

        params = tuple(params)
        bspline_transformation.SetParameters(tuple(params))

        for interpolator_idx, entry in enumerate(self.entries):
            data = sample[entry]
            for channel in range(data.shape[-1]):
                img = sitk.GetImageFromArray(data[..., channel])
                resampler = sitk.ResampleImageFilter()
                resampler.SetReferenceImage(img)
                resampler.SetInterpolator(self.interpolators[interpolator_idx])
                resampler.SetDefaultPixelValue(self.fill_value)
                resampler.SetTransform(bspline_transformation)

                img_deformed = resampler.Execute(img)
                sample[entry][..., channel] = sitk.GetArrayFromImage(img_deformed)

        return sample


class RandomMirror(tfm.Transform):

    def __init__(self, axis: int=-2, p: float=1.0, entries=('images', 'labels')):
        """Randomly mirrors the sample along a given axis.

        Args:
            p (float): The probability of the mirroring to be applied.
            axis (int): The axis to apply the mirroring.
            entries (tuple): The sample's entries to apply the mirroring to.
        """
        super().__init__()
        self.axis = axis
        self.p = p
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        if self.p < np.random.random():
            return sample

        for entry in self.entries:
            if entry not in sample:
                raise ValueError(tfm.ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))

            sample[entry] = np.flip(sample[entry], self.axis).copy()

        return sample


class RandomRotation90(tfm.Transform):

    def __init__(self, axes: typing.Tuple[int]=(-3, -2), p: float=1.0, entries=('images', 'labels')):
        """Randomly rotates the sample 90, 180, or 270 degrees in the plane specified by axes.

        Raises:
            UserWarning: If the plane to rotate is not rectangular.

        Args:
            axes (tuple): The sample is rotated in the plane defined by the axes.
                Axes must be of length two and different.
            p (float): The probability of the rotation to be applied.
            entries (tuple): The sample's entries to apply the rotation to.
        """
        super().__init__()
        if len(axes) !=2:
            raise ValueError('axes must be of length two')

        self.axes = axes
        self.p = p
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        if self.p < np.random.random():
            return sample

        k = np.random.randint(1, 4)

        for entry in self.entries:
            if entry not in sample:
                raise ValueError(tfm.ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))

            if sample[entry].shape[self.axes[0]] != sample[entry].shape[self.axes[1]]:
                raise UserWarning('entry "{}" has unequal in-plane dimensions ({}, {}). \
                Random 90 degree rotations might produce undesired results.'.format(entry,
                                                                                    sample[entry].shape[self.axes[0]],
                                                                                    sample[entry].shape[self.axes[1]]))
            sample[entry] = np.rot90(sample[entry], k, axes=self.axes).copy()

        return sample


class RandomShift(tfm.Transform):

    def __init__(self, shift: typing.Union[int, tuple], axis: typing.Union[int, tuple]=None,
                 mode: str='mirror', fill: float=0.0,
                 p: float=1.0, entries=('images', 'labels')):
        """Randomly shifts the sample along axes by a value from the interval [-p * size(axis), +p * size(axis)],
        where p is the percentage of shifting and size(axis) is the size along an axis.

        Args:
            shift (int, tuple): The percentage of shifting of the axis' size.
                If axis is not defined, the shifting will be applied from the first dimension onwards of the sample.
                Use None to exclude an axis or define axis to specify the axis/axes to crop.
                E.g.:
                    shift=0.2 with the default axis parameter shifts the sample along the 1st axis.
                    shift=(0.2, 0.1) with the default axis parameter shifts the sample along the 1st and 2nd axes.
                    shift=(None, 0.2) with the default axis parameter shifts the sample along the 2st axis.
                    shift=(0.2, 0.1) with axis=(1, 0) shifts the sample along the 1st and 2nd axes.
                    shift=(None, 0.1, 0.2) with axis=(1, 2, 0) shifts the sample along the 1st and 3rd axes.
            axis (int, tuple): Axis or axes to which the shift int or tuple correspond(s) to.
                If defined, must have the same length as shape.
            p (float): The probability of the shift to be applied.
            entries (tuple): The sample's entries to apply the shifting to.
        """
        super().__init__()
        if isinstance(shift, int):
            shape = (shift, )

        if axis is None:
            axis = tuple(range(len(shift)))
        if isinstance(axis, int):
            axis = (axis, )

        if len(axis) != len(shift):
            raise ValueError('If specified, the axis parameter must be of the same length as the shift')

        # filter out any axis where shift is None
        self.axis = tuple([a for a, s in zip(axis, shift) if s is not None])
        self.shift = tuple([s for s in shift if s is not None])

        self.p = p
        self.entries = entries

    def __call__(self, sample: dict) -> dict:
        if self.p < np.random.random():
            return sample

        for entry in self.entries:
            if entry not in sample:
                raise ValueError(tfm.ENTRY_NOT_EXTRACTED_ERR_MSG.format(entry))

        shifts_maximums = [int(s * sample[self.entries[0]].shape[a]) for a, s in zip(self.axis, self.shift)]
        shifts = [np.random.randint(-s_max, s_max) if s_max is not 0 else 0 for s_max in shifts_maximums]

        for entry in self.entries:
            for axis, shift in zip(self.axis, shifts):
                sample[entry] = np.roll(sample[entry], shift, axis)
                # todo(fabianbalsiger): implement zero filling (as optional "mode" parameter)?

        return sample



"""The pymia.evaluation.metric.regression module provides several metrics to measure regression performance."""

import numpy as np

from .base import INumpyArrayMetric


class MeanAbsoluteError(INumpyArrayMetric):
    """Represents a mean absolute error metric."""

    def __init__(self, metric: str = 'MAE'):
        """Initializes a new instance of the MeanAbsoluteError class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the mean absolute error."""

        return np.mean(np.abs(self.ground_truth - self.segmentation))


class MeanSquaredError(INumpyArrayMetric):
    """Represents a mean squared error metric."""

    def __init__(self, metric: str = 'MSE'):
        """Initializes a new instance of the MeanSquaredError class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the mean squared error."""

        return np.mean(np.square(self.ground_truth - self.segmentation))


class RootMeanSquaredError(INumpyArrayMetric):
    """Represents a root mean squared error metric."""

    def __init__(self, metric: str = 'RMSE'):
        """Initializes a new instance of the RootMeanSquaredError class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the root mean squared error."""

        return np.sqrt(np.mean(np.square(self.ground_truth - self.segmentation)))


class NormalizedRootMeanSquaredError(INumpyArrayMetric):
    """Represents a normalized root mean squared error metric."""

    def __init__(self, metric: str = 'NRMSE'):
        """Initializes a new instance of the NormalizedRootMeanSquaredError class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the normalized root mean squared error."""

        rmse = np.sqrt(np.mean(np.square(self.ground_truth - self.segmentation)))
        return rmse / (self.ground_truth.max() - self.ground_truth.min())


class CoefficientOfDetermination(INumpyArrayMetric):
    """Represents a coefficient of determination (R^2) error metric."""

    def __init__(self, metric: str = 'R2'):
        """Initializes a new instance of the CoefficientOfDetermination class.

        Args:
            metric (str): The identification string of the metric.
        """
        super().__init__(metric)

    def calculate(self):
        """Calculates the coefficient of determination (R^2) error.

        See Also:
            https://stackoverflow.com/a/45538060
        """
        y_true = self.ground_truth.flatten()
        y_predicted = self.segmentation.flatten()

        sse = sum((y_true - y_predicted) ** 2)
        tse = (len(y_true) - 1) * np.var(y_true, ddof=1)
        r2_score = 1 - (sse / tse)
        return r2_score



import abc
import json
import os

import yaml


class Dictable(metaclass=abc.ABCMeta):
    """Interface allowing dict serialization of an object."""

    @abc.abstractmethod
    def to_dict(self, **kwargs):
        """
        Transform instance to a dict

        Args:
            **kwargs: Additional arguments

        Returns:
            dict: Instance variables in dict
        """
        pass

    @abc.abstractmethod
    def from_dict(self, d: dict, **kwargs):
        """
        Updated instance by a dict

        Args:
            d(dict): Dict containing variables to be updated
            **kwargs: Additional arguments
        """
        pass


class ConfigurationBase(Dictable, metaclass=abc.ABCMeta):
    """Base class for Configurations."""
    @classmethod
    @abc.abstractmethod
    def version(cls) -> int:
        """Configuration version property"""
        pass

    @classmethod
    @abc.abstractmethod
    def type(cls) -> str:
        """Configuration type property"""
        pass

    def to_dict(self, **kwargs):
        return member_to_dict(self)

    def from_dict(self, d: dict, **kwargs):
        version = kwargs.get('version', self.version())
        self.handle_version(version, d)

        dict_to_member(self, d)

    def handle_version(self, version, d: dict):
        """ Version handling. To be overwritten if version has to be verified.

        Args:
            version(int): The version
            d: The configuration dict
        """
        pass

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        dict_as_string = str(self.to_dict())
        dict_as_string = dict_as_string.replace('{', '').replace('}', '').replace(',', '\n')
        return '{}:\n' \
               ' {}\n' \
            .format(type(self).__name__, dict_as_string)


def combine_dicts(target, source):
    for key, value in source.items():
        target[key] = value


def dict_to_member(target_obj, d: dict):
    members = vars(target_obj)  # vars has direct reference to writable members
    for key in members:
        if key in d:
            if isinstance(members[key], dict):
                combine_dicts(members[key], d[key])
            elif isinstance(members[key], Dictable):
                members[key].from_dict(d[key])
            else:
                members[key] = d[key]


def member_to_dict(target_obj):
    d = {}
    for k, v in vars(target_obj).items():
        if isinstance(v, Dictable):
            d[k] = v.to_dict()
        else:
            d[k] = v
    return d


class MetaData(Dictable):
    """Represents the meta-information in a configuration file."""
    def __init__(self, version: int=-1, type_: str='') -> None:
        self.version = version
        self.type = type_

    def to_dict(self, **kwargs):
        return vars(self)

    def from_dict(self, d: dict, **kwargs):
        self.version = d.get('version')
        self.type = d.get('type')


class ConfigurationParser(metaclass=abc.ABCMeta):
    """Represents the configuration parser interface."""

    @classmethod
    @abc.abstractmethod
    def load(cls, file_path: str, config_cls):
        """Load the configuration file.

        Args:
            file_path(str): The configuration file path
            config_cls: The type of the configuration

        Returns:
            ConfigurationBase: The configuration
        """
        pass

    @classmethod
    @abc.abstractmethod
    def save(cls, file_path: str, config: ConfigurationBase):
        """Save the configuration to file.

        Args:
            file_path: The configuration file path
            config: The configuration
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def read(file_path):
        """ Read the file content

        Args:
            file_path: The configuration file path

        Returns:
            The file content in the format of the underlying reader

        """

    @staticmethod
    @abc.abstractmethod
    def write(file_path, content):
        """Write the content file content

        Args:
            file_path: The configuration file path
            content: The content to write
        """


class JSONConfigurationParser(ConfigurationParser):
    """Represents a JSON configuration parser."""

    @classmethod
    def load(cls, file_path: str, config_cls):
        if not os.path.isfile(file_path):
            raise ValueError('File {} does not exist'.format(file_path))

        d = cls.read(file_path)

        meta_dict = d['meta']
        config_dict = d['config']

        meta = MetaData()
        meta.from_dict(meta_dict)

        config = config_cls()
        if meta.type != config.type():
            raise ValueError('configuration type "{}" does not fit the configuration instance "{}"'.format(
                meta.type, config.__class__.__name__))

        config.from_dict(config_dict, version=meta.version)
        return config

    @classmethod
    def save(cls, file_path: str, config: ConfigurationBase):
        meta = MetaData(config.version(), config.type())
        d = {'meta': meta.to_dict(), 'config': config.to_dict()}

        cls.write(file_path, d)

    @staticmethod
    def read(file_path):
        with open(file_path, 'r') as f:
            d = json.load(f)
        return d

    @staticmethod
    def write(file_path, content):
        with open(file_path, 'w') as f:
            json.dump(content, f, sort_keys=True, indent=4)


class YamlConfigurationParser(JSONConfigurationParser):
    """Represents a YAML configuration parser."""

    @staticmethod
    def read(file_path):
        with open(file_path, 'r') as f:
            d = yaml.load(f)
        return d

    @staticmethod
    def write(file_path, content):
        with open(file_path, 'w') as f:
            yaml.dump(content, f, default_flow_style=False)


def load(file_path: str, config_cls) -> ConfigurationBase:
    """Load the configuration file.

    Args:
        file_path(str): The configuration file path
        config_cls: The configuration class type

    Returns:
        ConfigurationBase: The configuration
    """
    extension = os.path.splitext(file_path)[1]
    if extension not in parser_registry:
        raise ValueError('unknown configuration file extension "{}"'.format(extension))

    return parser_registry[extension].load(file_path, config_cls)


def save(file_path: str, config: ConfigurationBase) -> None:
    """Save the configuration to file.

    Args:
        file_path: The configuration file path
        config: The configuration
    """

    extension = os.path.splitext(file_path)[1]
    if extension not in parser_registry:
        raise ValueError('unknown configuration file extension "{}"'.format(extension))

    return parser_registry[extension].save(file_path, config)


def load_meta(file_path: str) -> MetaData:
    """ Load the meta data of a configuration

    Args:
        file_path: The configuration file path

    Returns:
        MetaData: The meta data of the configuration
    """
    extension = os.path.splitext(file_path)[1]
    if extension not in parser_registry:
        raise ValueError('unknown configuration file extension "{}"'.format(extension))

    d = parser_registry[extension].read(file_path)
    meta_data = MetaData()
    meta_data.from_dict(d['meta'])
    return meta_data


parser_registry = {'.json': JSONConfigurationParser, '.yaml': YamlConfigurationParser, '.yml': YamlConfigurationParser}



try:
    import torch
except ImportError:
    raise ImportError('Module "torch" not found. Install "torch==1.3.1" to use the data module.')

from .indexexpression import IndexExpression
from .subjectfile import (SubjectFile)



# https://github.com/aicodes/tf-bestpractice
# https://github.com/MrGemy95/Tensorflow-Project-Template
import abc
import glob
import logging
import os
import re

import numpy as np
import tensorflow as tf
import torch


class Model(abc.ABC):

    def __init__(self):
        self.loss = None  # The loss function
        self.optimizer = None  # The optimizer
        self.network = None  # The network itself

    @abc.abstractmethod
    def inference(self, x) -> object:
        """This is the network, i.e. the forward calculation from x to y."""
        # return some_op(x, name="inference")
        pass

    @abc.abstractmethod
    def loss_function(self, prediction, label=None, **kwargs):
        pass

    @abc.abstractmethod
    def optimize(self, **kwargs):
        pass

    @abc.abstractmethod
    def save(self, path: str, epoch: int, **kwargs):
        """Saves the model."""
        pass

    @abc.abstractmethod
    def load(self, path: str) -> bool:
        """Loads a model.

        Args:
            path: The path to the model.

        Returns:
            True, if the model was loaded; otherwise, False.
        """
        pass

    @abc.abstractmethod
    def set_epoch(self, epoch: int):
        pass

    @abc.abstractmethod
    def get_number_parameters(self) -> int:
        """Gets the number of trainable parameters of the model.

        Returns:
            The number of parameters.
        """
        pass


class TensorFlowModel(Model, abc.ABC):

    def __init__(self, session, model_dir: str, max_to_keep: int=3, **kwargs):
        super().__init__()

        self.session = session
        self.model_dir = model_dir

        # initialize placeholders
        self.x_placeholder = None
        self.y_placeholder = None
        self.is_training_placeholder = None
        self.placeholders(kwargs['x_shape'], kwargs['y_shape'])

        # initialize global step, i.e. the number of batches seen by the graph (starts at 0)
        self.global_step = tf.train.get_or_create_global_step()  # todo: with tf.variable_scope('global_step')
        # initialize epoch, i.e. the number of epochs trained (starts at 1)
        self.epoch = tf.Variable(1, name='epoch', trainable=False)
        self.epoch_placeholder = tf.placeholder(self.epoch.dtype)
        self.epoch_op = self.epoch.assign(self.epoch_placeholder)
        # initialize best model score
        self.best_model_score = tf.Variable(0.0, name='best_model_score', trainable=False)
        self.best_model_score_placeholder = tf.placeholder(self.best_model_score.dtype)
        self.best_model_score_op = self.best_model_score.assign(self.best_model_score_placeholder)
        # self.best_model_score_name = tf.Variable('unknown score', name='best_model_score_name', trainable=False)
        # self.best_model_score_name_placeholder = tf.placeholder(self.best_model_score_name.dtype)
        # self.best_model_score_name_op = self.best_model_score_name.assign(self.best_model_score_name_placeholder)

        self.network = self.inference(self.x_placeholder)
        self.loss = self.loss_function(self.network, self.y_placeholder)
        self.optimizer = self.optimize()

        self.saver = tf.train.Saver(max_to_keep=max_to_keep)
        self.saver_best = tf.train.Saver(max_to_keep=1)

    @abc.abstractmethod
    def placeholders(self, x_shape: tuple, y_shape: tuple):
        raise NotImplementedError('placeholders')

    def save(self, path: str, epoch: int, **kwargs):
        if 'best_model_score' in kwargs:
            # update best model score variable in graph (such that it gets automatically restored
            score = kwargs['best_model_score']
            self.session.run(self.best_model_score_op, feed_dict={self.best_model_score_placeholder: score})

            if 'best_model_score_name' in kwargs:
                score_name = kwargs['best_model_score_name']
            else:
                score_name = 'score'

            #  self.session.run(self.best_model_score_name_op,
            #                   feed_dict={self.best_model_score_name_placeholder: score_name})

            # we save with a different checkpoint file, such that max_to_keep applies only to the epoch savings
            # and the best model does not get overridden
            saved_checkpoint = self.saver_best.save(self.session, path, latest_filename='checkpoint_best')
            logging.info('Epoch {:d}: Saved best model with {} of {:.6f} at {}'.format(
                epoch, score_name, score, saved_checkpoint))
        else:
            saved_checkpoint = self.saver.save(self.session, path, global_step=epoch)
            logging.info('Epoch {:d}: Saved model at {}'.format(epoch, saved_checkpoint))

    def load(self, path: str) -> bool:
        # TensorFlow automatically restores self.best_model_score, self.epoch, self.global_step etc.
        # check e.g. with self.epoch.eval()
        latest_checkpoint = tf.train.latest_checkpoint(path)
        if latest_checkpoint:
            self.saver.restore(self.session, latest_checkpoint)
            logging.info('Loaded model from {}'.format(latest_checkpoint))
            return True
        elif not os.path.isdir(path):
            # try to load model directly
            self.saver.restore(self.session, path)
            logging.info('Loaded model from {}'.format(path))
            return True
        else:
            return False

    def set_epoch(self, epoch: int):
        self.session.run(self.epoch_op, feed_dict={self.epoch_placeholder: epoch})

    def get_number_parameters(self) -> int:
        no_parameters = np.sum([np.prod(v.shape.as_list()) for v in tf.trainable_variables()])
        return int(no_parameters)


class TorchModel(Model, abc.ABC):

    def __init__(self, model_dir: str, max_to_keep: int=3):
        super().__init__()

        self.model_dir = model_dir
        self.max_to_keep = max_to_keep

        # initialize global step, i.e. the number of batches seen by the graph (starts at 0)
        # this is TensorFlow notation but required since we use TensorBoard. It can be seen as the logging step.
        self.global_step = 0
        # initialize epoch, i.e. the number of epochs trained (starts at 1)
        self.epoch = 1
        # initialize best model score
        self.best_model_score = 0
        self.best_model_score_name = 'unknown score'

        self.device = None

    def save(self, path: str, epoch: int, **kwargs):
        if 'best_model_score' in kwargs:
            self.best_model_score = kwargs['best_model_score']
            if 'best_model_score_name' in kwargs:
                self.best_model_score_name = kwargs['best_model_score_name']
            save_path = path + '.pt'
            logging_str = 'Epoch {:d}: Saved best model with {} of {:.6f} at {}'.format(
                epoch, self.best_model_score_name, self.best_model_score, save_path)
        else:
            save_path = path + '-{}.pt'.format(epoch)
            logging_str = 'Epoch {:d}: Saved model at {}'.format(epoch, save_path)

        state_dict = {
            'best_model_score': self.best_model_score,
            'best_model_score_name': self.best_model_score_name,
            'epoch': epoch,
            'global_step': self.global_step,  # todo(fabianbalsiger): is always zero
            'network_state_dict': self.network.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
        }

        torch.save(state_dict, save_path)
        logging.info(logging_str)

        # keep only max_to_keep models
        saved_models = glob.glob(path + '-[0-9]*.pt')
        saved_models = sorted(saved_models,
                              key=lambda model_path: int(re.search(path + '-(\d+).pt', model_path).group(1)))

        for saved_model in saved_models[:-self.max_to_keep]:
            os.remove(saved_model)

    def load(self, path: str) -> bool:
        if os.path.isfile(path):
            state_dict = torch.load(path)
        else:
            # check if a saved model is available
            saved_models = glob.glob(os.path.join(path, '*-[0-9]*.pt'))
            if len(saved_models) == 0:
                return False
            saved_models = sorted(saved_models,
                                  key=lambda model_path: int(re.search('-(\d+).pt', model_path).group(1)))
            state_dict = torch.load(saved_models[-1])

        self.best_model_score = state_dict['best_model_score']
        if 'best_model_score_name' in state_dict:
            # todo(fabianbalsiger): remove condition in next major release (which will brake backwards compatibility)?
            self.best_model_score_name = state_dict['best_model_score_name']
        self.epoch = state_dict['epoch']
        self.global_step = state_dict['global_step']
        self.network.load_state_dict(state_dict['network_state_dict'])
        self.optimizer.load_state_dict(state_dict['optimizer_state_dict'])

        return True

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def get_number_parameters(self) -> int:
        no_parameters = np.sum([np.prod(p.size()) for p in self.network.parameters() if p.requires_grad])
        return int(no_parameters)



from .base import (ConfusionMatrix, Distances, IMetric, IConfusionMatrixMetric, IDistanceMetric, ISimpleITKImageMetric,
                   INumpyArrayMetric, Information, NotComputableMetricWarning)
from .metric import (get_all_segmentation_metrics, get_all_regression_metrics, get_overlap_metrics, get_distance_metrics,
                     get_classical_metrics)

from .regression import (CoefficientOfDetermination, MeanAbsoluteError, MeanSquaredError, RootMeanSquaredError,
                         NormalizedRootMeanSquaredError)
from .segmentation import (Accuracy, AdjustedRandIndex, AreaUnderCurve, AverageDistance, CohenKappaMetric,
                           DiceCoefficient, FalseNegative, FalsePositive, Fallout, FalseNegativeRate, FMeasure,
                           GlobalConsistencyError, GroundTruthArea, GroundTruthVolume, HausdorffDistance,
                           InterclassCorrelation, JaccardCoefficient, MahalanobisDistance, MutualInformation, Precision,
                           ProbabilisticDistance, RandIndex, Recall, SegmentationArea, SegmentationVolume,
                           Sensitivity, Specificity, SurfaceOverlap, SurfaceDiceOverlap, TrueNegative, TruePositive,
                           VariationOfInformation, VolumeSimilarity)



"""This module holds classes related to validation."""


class LeaveOneOutCrossValidator:
    """Represents a leave-one-out cross-validation."""

    def __init__(self, n: int):
        """Initializes a new instance of the LeaveOneOutCrossValidator class.

        Args:
            n (int): The number of samples.
        """
        self.i = 0
        self.n = n

    def __iter__(self):
        """Gets an iterator object.

        Returns:
            LeaveOneOutCrossValidator: Itself.
        """
        return self

    def __next__(self):
        """Generates the next training and testing indices.

        Returns:
            (list of int, list of int): The training and testing indices, respectively.
        """
        if self.i < self.n:
            test = self.i
            train = list(range(self.n))
            del train[test]
            self.i += 1
            return train, test
        else:
            raise StopIteration()

    def __str__(self):
        """Gets a printable string representation.

        Returns:
            str: String representation.
        """
        return 'LeaveOneOutCrossValidator:\n' \
               ' i: {self.i}\n' \
               ' n: {self.n}\n' \
            .format(self=self)



class FileCategory:
    def __init__(self, entries=None) -> None:
        if entries is None:
            entries = {}
        self.entries = entries


class SubjectFile:

    def __init__(self, subject: str, **file_groups) -> None:
        self.subject = subject
        self.categories = {}
        for category_name, category_files in file_groups.items():
            self.categories[category_name] = FileCategory(category_files)
        self._check_validity()

    def _check_validity(self):
        all_file_ids = []
        for file_category in self.categories.values():
            all_file_ids.extend(file_category.entries.keys())

        if len(all_file_ids) > len(set(all_file_ids)):
            raise ValueError('Identifiers must be unique')

    def get_all_files(self):
        all_files = {}
        for file_category in self.categories.values():
            for id_, filename in file_category.entries.items():
                all_files[id_] = filename
        return all_files



import os
import sys
from setuptools import find_packages, setup

if sys.version_info < (3, 6):
    sys.exit("Requires Python 3.6 or higher")

directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join('README.rst'), encoding='utf-8') as f:
    readme = f.read()

REQUIRED_PACKAGES = [
    'h5py >= 2.10.0',
    'matplotlib >= 3.1.1',
    'numpy >= 1.12.1',
    'SimpleITK >= 1.2.3',
    'pyyaml >= 5.1.2',
    'scipy >= 1.3.2'
]

TEST_PACKAGES = [

]

setup(
    name='pymia',
    version='0.2.2',
    description='pymia facilitates medical image analysis',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Fabian Balsiger and Alain Jungo',
    author_email='fabian.balsiger@artorg.unibe.ch',
    url='https://github.com/rundherum/pymia',
    license='Apache License',
    python_requires='>=3.6',
    packages=find_packages(exclude=['docs', 'examples', 'test']),
    install_requires=REQUIRED_PACKAGES,
    tests_require=REQUIRED_PACKAGES + TEST_PACKAGES,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries'
    ],
    keywords=[
        'medical image analysis',
        'deep learning',
        'data handling',
        'evaluation',
        'metrics'
    ]
)



import abc
import pickle
import typing as t

import numpy as np
import SimpleITK as sitk

import pymia.data.conversion as conv
import pymia.data.definition as df
import pymia.data.indexexpression as expr
from . import reader as rd


class Extractor(metaclass=abc.ABCMeta):
    """Represents an extractor that extracts data from a dataset."""

    @abc.abstractmethod
    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        pass


class ComposeExtractor(Extractor):
    """Composes multiple Extractor objects."""

    def __init__(self, extractors) -> None:
        super().__init__()
        self.extractors = extractors

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        for e in self.extractors:
            e.extract(reader, params, extracted)


class NamesExtractor(Extractor):
    """Extracts the names of the entries within a category (i.e. the name of the enum identifying the data).

    The names are of type str.
    """

    def __init__(self, cache: bool=True, categories=('images', 'labels')) -> None:
        super().__init__()
        self.cache = cache
        self.cached_result = None
        self.categories = categories

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        if not self.cache or self.cached_result is None:
            d = self._extract(reader)
            self.cached_result = d
        else:
            d = self.cached_result

        for k, v in d.items():
            extracted[k] = v

    def _extract(self, reader: rd.Reader):
        d = {}
        for category in self.categories:
            d['{}_names'.format(category)] = reader.read(df.NAMES_PLACEHOLDER.format(category))
        return d


class SubjectExtractor(Extractor):
    """Extracts the subject's identification.

    The subject's identification is of type str.
    """

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        extracted['subject_index'] = params['subject_index']
        subject_index_expr = expr.IndexExpression(params['subject_index'])
        extracted['subject'] = reader.read(df.SUBJECT, subject_index_expr)


class IndexingExtractor(Extractor):
    """Extracts the index expression.

    The index expression is of type IndexExpression.
    """

    def __init__(self, do_pickle: bool=False) -> None:
        super().__init__()
        self.do_pickle = do_pickle

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        extracted['subject_index'] = params['subject_index']
        index_expression = params['index_expr']
        if self.do_pickle:
            # pickle to prevent from problems since own class
            index_expression = pickle.dumps(index_expression)
        extracted['index_expr'] = index_expression


class ImagePropertiesExtractor(Extractor):
    """Extracts the image properties.

    The image properties are of type ImageProperties.
    """

    def __init__(self, do_pickle: bool=False) -> None:
        super().__init__()
        self.do_pickle = do_pickle

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        subject_index_expr = expr.IndexExpression(params['subject_index'])

        shape = reader.read(df.INFO_SHAPE, subject_index_expr).tolist()
        direction = reader.read(df.INFO_DIRECTION, subject_index_expr).tolist()
        spacing = reader.read(df.INFO_SPACING, subject_index_expr).tolist()
        origin = reader.read(df.INFO_ORIGIN, subject_index_expr).tolist()

        # todo: everything in memory?
        image = sitk.Image(shape, sitk.sitkUInt8)
        image.SetDirection(direction)
        image.SetSpacing(spacing)
        image.SetOrigin(origin)
        # todo number_of_components_per_pixel and pixel_id

        img_properties = conv.ImageProperties(image)
        if self.do_pickle:
            # pickle to prevent from problems since own class
            img_properties = pickle.dumps(img_properties)
        extracted['properties'] = img_properties


class FilesExtractor(Extractor):
    """Extracts the file paths.

    The file paths are of type str.
    """

    def __init__(self, cache: bool=True, categories=('images', 'labels')) -> None:
        super().__init__()
        self.cache = cache
        self.cached_file_root = None
        self.categories = categories

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        subject_index_expr = expr.IndexExpression(params['subject_index'])

        if not self.cache or self.cached_file_root is None:
            file_root = reader.read(df.FILES_ROOT)
            self.cached_file_root = file_root
        else:
            file_root = self.cached_file_root

        extracted['file_root'] = file_root

        for category in self.categories:
            extracted['{}_files'.format(category)] = reader.read(df.FILES_PLACEHOLDER.format(category),
                                                                 subject_index_expr)


class SelectiveDataExtractor(Extractor):
    """Extracts data of a given category selectively."""

    def __init__(self, selection=None, category: str='labels') -> None:
        """Initializes a new instance of the SelectiveDataExtractor class.

        Args:
            selection (str or tuple): Entries within the category to select.
                If selection is None, the class has the same behaviour as the DataExtractor and selects all entries.
            category (str): The category to extract data from.
        """
        super().__init__()
        self.entry_base_names = None

        if isinstance(selection, str):
            selection = (selection,)
        self.selection = selection
        self.category = category

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        if '{}_names'.format(self.category) not in extracted:
            raise ValueError('selection of labels requires label_names to be extracted (use NamesExtractor)')

        if self.entry_base_names is None:
            entries = reader.get_subject_entries()
            self.entry_base_names = [entry.rsplit('/', maxsplit=1)[1] for entry in entries]

        if not reader.has(df.DATA_PLACEHOLDER.format(self.category)):
            raise ValueError('SelectiveDataExtractor requires {} to exist'.format(self.category))

        subject_index = params['subject_index']
        index_expr = params['index_expr']

        base_name = self.entry_base_names[subject_index]
        data = reader.read('{}/{}'.format(df.DATA_PLACEHOLDER.format(self.category), base_name), index_expr)
        label_names = extracted['{}_names'.format(self.category)]  # type: list

        if self.selection is None:
            extracted[self.category] = data
        else:
            selection_indices = np.array([label_names.index(s) for s in self.selection])
            extracted[self.category] = np.take(data, selection_indices, axis=-1)


class RandomDataExtractor(Extractor):
    """Extracts data of a given category randomly."""

    def __init__(self, selection=None, category: str='labels') -> None:
        """Initializes a new instance of the RandomDataExtractor class.

        Args:
            selection (str or tuple): Entries within the category to select an entry randomly from.
                If selection is None, an entry from all entries is randomly selected.
            selection (str or tuple): Note that if selection is None, all keys are considered.
            category (str): The category to extract data from.
        """
        super().__init__()
        self.entry_base_names = None

        if isinstance(selection, str):
            selection = (selection,)
        self.selection = selection
        self.category = category

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        if '{}_names'.format(self.category) not in extracted:
            raise ValueError('selection of labels requires label_names to be extracted (use NamesExtractor)')

        if self.entry_base_names is None:
            entries = reader.get_subject_entries()
            self.entry_base_names = [entry.rsplit('/', maxsplit=1)[1] for entry in entries]

        if not reader.has(df.DATA_PLACEHOLDER.format(self.category)):
            raise ValueError('SelectiveDataExtractor requires {} to exist'.format(self.category))

        subject_index = params['subject_index']
        index_expr = params['index_expr']

        base_name = self.entry_base_names[subject_index]
        data = reader.read('{}/{}'.format(df.DATA_PLACEHOLDER.format(self.category), base_name), index_expr)
        label_names = extracted['{}_names'.format(self.category)]  # type: list

        if self.selection is None:
            selection_indices = np.array(range(len(label_names)))
        else:
            selection_indices = np.array([label_names.index(s) for s in self.selection])

        random_index = [np.random.choice(selection_indices)]  # as list to keep the last dimension with np.take
        extracted[self.category] = np.take(data, random_index, axis=-1)


class ImageShapeExtractor(Extractor):

    def __init__(self, numpy_format: bool=True) -> None:
        super().__init__()
        self.numpy_format = numpy_format

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        subject_index_expr = expr.IndexExpression(params['subject_index'])

        shape = reader.read(df.INFO_SHAPE, subject_index_expr)
        if self.numpy_format:
            tmp = shape[0]
            shape[0] = shape[-1]
            shape[-1] = tmp

        extracted['shape'] = tuple(shape.tolist())


class DataExtractor(Extractor):

    def __init__(self, categories=('images',), ignore_indexing: bool=False) -> None:
        super().__init__()
        self.categories = categories
        self.ignore_indexing = ignore_indexing
        self.entry_base_names = None

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        if self.entry_base_names is None:
            entries = reader.get_subject_entries()
            self.entry_base_names = [entry.rsplit('/', maxsplit=1)[1] for entry in entries]

        subject_index = params['subject_index']
        index_expr = params['index_expr']

        base_name = self.entry_base_names[subject_index]
        for category in self.categories:
            if self.ignore_indexing:
                data = reader.read('{}/{}'.format(df.DATA_PLACEHOLDER.format(category), base_name))
            else:
                data = reader.read('{}/{}'.format(df.DATA_PLACEHOLDER.format(category), base_name), index_expr)
            extracted[category] = data


class PadDataExtractor(Extractor):

    def __init__(self, padding: t.Union[tuple, t.List[tuple]], extractor: Extractor, pad_fn=None):
        super().__init__()
        if not (hasattr(extractor, 'categories') or hasattr(extractor, 'category')):
            raise ValueError('argument extractor needs to have the property "categories" or "category"')

        self.extractor = extractor
        self.pad_fn = pad_fn

        if self.pad_fn is None:
            self.pad_fn = PadDataExtractor.zero_pad

        if isinstance(padding[0], int):
            padding = [(pad, pad) for pad in padding]
        index_diffs = np.asarray(padding)
        index_diffs[:, 0] = -index_diffs[:, 0]
        self.index_diffs = index_diffs

    def extract(self, reader: rd.Reader, params: dict, extracted: dict) -> None:
        index_expr = params['index_expr']  # type: expr.IndexExpression

        # Make sure all indexing is done with slices (Example: (16,) will be changed to (slice(16, 17, None),) which
        # is equivalent), otherwise the following steps will be wrong; .
        if any(isinstance(s, int) for s in index_expr.expression):
            index_expr.set_indexing([slice(s, s + 1) if isinstance(s, int) else s for s in index_expr.expression])

        padded_indexing = np.asarray(index_expr.get_indexing()) + self.index_diffs
        padded_shape = tuple((padded_indexing[:, 1] - padded_indexing[:, 0]).tolist())

        sub_indexing = padded_indexing.copy()
        sub_indexing[padded_indexing > 0] = 0
        sub_indexing = -sub_indexing

        padded_indexing[padded_indexing < 0] = 0  # cannot slice outside the boundary in negative (but positive works!)
        padded_index_expr = expr.IndexExpression(padded_indexing.tolist())

        padded_params = params.copy()
        padded_params['index_expr'] = padded_index_expr
        self.extractor.extract(reader, padded_params, extracted)

        categories = self.extractor.categories if hasattr(self.extractor, 'categories') else [self.extractor.category]

        for category in categories:
            data = extracted[category]

            full_pad_shape = padded_shape + data.shape[len(padded_shape):]
            if full_pad_shape != data.shape:
                # we could not fully extract the padded shape, use pad_fn to pad data
                extracted[category] = self.pad_fn(data, full_pad_shape, sub_indexing)

    @staticmethod
    def zero_pad(data: np.ndarray, pad_shape, sub_indexing):
        pad_data = np.zeros(pad_shape, dtype=data.dtype)

        sub_indexing[:, 1] = sub_indexing[:, 0] + data.shape[:sub_indexing.shape[0]]
        sub_index_expr = expr.IndexExpression(sub_indexing.tolist())

        pad_data[sub_index_expr.expression] = data
        return pad_data



import unittest

import numpy as np
import SimpleITK as sitk

import pymia.data.conversion as img


class TestImageProperties(unittest.TestCase):
    def test_is_two_dimensional(self):
        x = 10
        y = 10
        image = sitk.Image([x, y], sitk.sitkUInt8)
        dut = img.ImageProperties(image)

        self.assertEqual(dut.is_two_dimensional(), True)
        self.assertEqual(dut.is_three_dimensional(), False)
        self.assertEqual(dut.is_vector_image(), False)

    def test_is_three_dimensional(self):
        x = 10
        y = 10
        z = 3
        image = sitk.Image([x, y, z], sitk.sitkUInt8)
        dut = img.ImageProperties(image)

        self.assertEqual(dut.is_two_dimensional(), False)
        self.assertEqual(dut.is_three_dimensional(), True)
        self.assertEqual(dut.is_vector_image(), False)

    def test_is_vector_image(self):
        x = 10
        y = 10
        number_of_components_per_pixel = 3
        image = sitk.Image([x, y], sitk.sitkVectorUInt8, number_of_components_per_pixel)
        dut = img.ImageProperties(image)

        self.assertEqual(dut.is_two_dimensional(), True)
        self.assertEqual(dut.is_three_dimensional(), False)
        self.assertEqual(dut.is_vector_image(), True)

    def test_properties(self):
        x = 10
        y = 10
        z = 3
        pixel_id = sitk.sitkUInt8
        size = (x, y, z)
        direction = (0, 1, 0, 1, 0, 0, 0, 0, 1)
        image = sitk.Image([x, y, z], pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut = img.ImageProperties(image)

        self.assertEqual(dut.size, size)
        self.assertEqual(dut.origin, size)
        self.assertEqual(dut.spacing, size)
        self.assertEqual(dut.direction, direction)
        self.assertEqual(dut.dimensions, z)
        self.assertEqual(dut.number_of_components_per_pixel, 1)
        self.assertEqual(dut.pixel_id, pixel_id)

    def test_equality(self):
        x = 10
        y = 10
        z = 3
        pixel_id = sitk.sitkUInt8
        size = (x, y, z)
        direction = (0, 1, 0, 1, 0, 0, 0, 0, 1)
        image = sitk.Image([x, y, z], pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut1 = img.ImageProperties(image)
        dut2 = img.ImageProperties(image)

        self.assertTrue(dut1 == dut2)
        self.assertFalse(dut1 != dut2)

        image = sitk.Image([x, y, z], sitk.sitkInt8)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut1 = img.ImageProperties(image)

        self.assertTrue(dut1 == dut2)
        self.assertFalse(dut1 != dut2)

        image = sitk.Image([x, y, z], sitk.sitkVectorUInt8, 2)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut1 = img.ImageProperties(image)

        self.assertTrue(dut1 == dut2)
        self.assertFalse(dut1 != dut2)

    def test_non_equality(self):
        x = 10
        y = 10
        z = 3
        pixel_id = sitk.sitkUInt8
        size = (x, y, z)
        direction = (0, 1, 0, 1, 0, 0, 0, 0, 1)
        image = sitk.Image([x, y, z], pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut1 = img.ImageProperties(image)

        different_size = (x, y, 2)

        # non-equal size
        image = sitk.Image(different_size, pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut2 = img.ImageProperties(image)

        self.assertFalse(dut1 == dut2)
        self.assertTrue(dut1 != dut2)

        # non-equal origin
        image = sitk.Image(size, pixel_id)
        image.SetOrigin(different_size)
        image.SetSpacing(size)
        image.SetDirection(direction)
        dut2 = img.ImageProperties(image)

        self.assertFalse(dut1 == dut2)
        self.assertTrue(dut1 != dut2)

        # non-equal spacing
        different_size = (x, y, 2)
        image = sitk.Image(size, pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(different_size)
        image.SetDirection(direction)
        dut2 = img.ImageProperties(image)

        self.assertFalse(dut1 == dut2)
        self.assertTrue(dut1 != dut2)

        # non-equal direction
        different_size = (x, y, 2)
        image = sitk.Image(size, pixel_id)
        image.SetOrigin(size)
        image.SetSpacing(size)
        image.SetDirection((1, 0, 0, 0, 1, 0, 0, 0, 1))
        dut2 = img.ImageProperties(image)

        self.assertFalse(dut1 == dut2)
        self.assertTrue(dut1 != dut2)


class TestNumpySimpleITKImageBridge(unittest.TestCase):

    def setUp(self):
        dim_x = 5
        dim_y = 10
        dim_z = 3
        self.no_vector_components = 4

        # some unordinary origins, spacings and directions
        self.origin_spacing_2d = (dim_x, dim_y)
        self.direction_2d = (0, 1, 1, 0)
        self.origin_spacing_3d = (dim_x, dim_y, dim_z)
        self.direction_3d = (1, 0, 0, 0, 1, 0, 0, 0, 1)

        # set up images
        image = sitk.Image((dim_x, dim_y, dim_z), sitk.sitkUInt8)
        image.SetOrigin(self.origin_spacing_3d)
        image.SetSpacing(self.origin_spacing_3d)
        image.SetDirection(self.direction_3d)
        self.properties_3d = img.ImageProperties(image)

        image = sitk.Image((dim_x, dim_y), sitk.sitkUInt8)
        image.SetOrigin(self.origin_spacing_2d)
        image.SetSpacing(self.origin_spacing_2d)
        image.SetDirection(self.direction_2d)
        self.properties_2d = img.ImageProperties(image)

        # set up numpy arrays (note the inverted order of the dimension)
        self.array_image_shape_2d = np.zeros((dim_y, dim_x), np.uint8)
        self.array_2d_vector = np.zeros((dim_y * dim_x, self.no_vector_components), np.uint8)
        self.array_image_shape_2d_vector = np.zeros((dim_y, dim_x, self.no_vector_components), np.uint8)

        self.array_image_shape_3d = np.zeros((dim_z, dim_y, dim_x), np.uint8)
        self.array_3d_vector = np.zeros((dim_z * dim_y * dim_x, self.no_vector_components), np.uint8)
        self.array_image_shape_3d_vector = np.zeros((dim_z, dim_y, dim_x, self.no_vector_components), np.uint8)

    def test_vector_to_image(self):
        # test array shape (n,) i.e., (x * y) and (x * y * z)
        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_3d.flatten(), self.properties_3d)
        self._assert_3d(image)

        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_2d.flatten(), self.properties_2d)
        self._assert_2d(image)

    def test_array_to_image(self):
        # test array shape (y, x) and (z, y, x)
        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_3d, self.properties_3d)
        self._assert_3d(image)

        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_2d, self.properties_2d)
        self._assert_2d(image)

    def test_array_to_vector_image(self):
        # test array shape (y, x, v) and (z, y, x, v), where v = number of vector component
        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_3d_vector, self.properties_3d)
        self._assert_3d(image, True)

        image = img.NumpySimpleITKImageBridge.convert(self.array_image_shape_2d_vector, self.properties_2d)
        self._assert_2d(image, True)

    def test_vector_to_vector_image(self):
        # test array shape (y * x, v) and (z * y * x, v), where v = number of vector components
        image = img.NumpySimpleITKImageBridge.convert(self.array_3d_vector, self.properties_3d)
        self._assert_3d(image, True)

        image = img.NumpySimpleITKImageBridge.convert(self.array_2d_vector, self.properties_2d)
        self._assert_2d(image, True)

    def test_convert_unknown_shape(self):
        with self.assertRaises(ValueError):
            img.NumpySimpleITKImageBridge.convert(self.array_image_shape_3d.flatten(), self.properties_2d)

    def _assert_2d(self, image: sitk.Image, is_vector=False):
        self.assertEqual(self.properties_2d.size, image.GetSize())

        if is_vector:
            self.assertEqual(self.no_vector_components, image.GetNumberOfComponentsPerPixel())
        else:
            self.assertEqual(1, image.GetNumberOfComponentsPerPixel())

        self.assertEqual(self.origin_spacing_2d, image.GetOrigin())
        self.assertEqual(self.origin_spacing_2d, image.GetSpacing())
        self.assertEqual(self.direction_2d, image.GetDirection())

    def _assert_3d(self, image: sitk.Image, is_vector=False):
        self.assertEqual(self.properties_3d.size, image.GetSize())

        if is_vector:
            self.assertEqual(self.no_vector_components, image.GetNumberOfComponentsPerPixel())
        else:
            self.assertEqual(1, image.GetNumberOfComponentsPerPixel())

        self.assertEqual(self.origin_spacing_3d, image.GetOrigin())
        self.assertEqual(self.origin_spacing_3d, image.GetSpacing())
        self.assertEqual(self.direction_3d, image.GetDirection())


class TestSimpleITKNumpyImageBridge(unittest.TestCase):
    def test_convert(self):
        x = 10
        y = 10
        z = 3
        size = (x, y, z)
        image = sitk.Image(size, sitk.sitkUInt8)

        array, properties = img.SimpleITKNumpyImageBridge.convert(image)

        self.assertEqual(isinstance(array, np.ndarray), True)
        self.assertEqual(array.shape, size[::-1])
        self.assertEqual(array.dtype, np.uint8)
        self.assertEqual(isinstance(properties, img.ImageProperties), True)
        self.assertEqual(properties.size, size)

    def test_convert_None(self):
        with self.assertRaises(ValueError):
            img.SimpleITKNumpyImageBridge.convert(None)



# see also https://gist.github.com/kukuruza/03731dc494603ceab0c5
import numpy as np


def get_grid_size(kernel_shape: list) -> tuple:
    size = int(np.ceil(np.sqrt(kernel_shape[-1])))
    space = 1  # the space between different kernels (in pixel)
    patch_h = kernel_shape[0]
    patch_w = kernel_shape[1]

    if patch_h == 1 and patch_w == 1:
        # 1-D conv
        w = kernel_shape[2]
        h = kernel_shape[3] + (kernel_shape[3] - 1) * space
    else:
        w = size * patch_w + (size - 1) * space
        h = size * patch_h + (size - 1) * space

    return h, w


def make_grid(kernels: np.ndarray) -> np.ndarray:
    """
    Arrange kernel weights on a grid for visualization
    :param kernels: N kernel (weights) as [N, h, w, 1]
    :return: 2D ndarray (gray-scale image)
    """
    size = int(np.ceil(np.sqrt(kernels.shape[0])))
    space = 1
    patch_h = kernels.shape[1]
    patch_w = kernels.shape[2]

    w = size * patch_w + (size-1) * space
    h = size * patch_h + (size - 1) * space

    grid = np.zeros([h, w])
    idx = 0
    for row in range(size):
        y = row * (patch_h + space)
        for col in range(size):
            x = col * (patch_w + space)
            if idx < kernels.shape[0]:
                patch = kernels[idx, :, :, 0]

                # normalize patch
                patch = (patch - np.max(patch)) / -np.ptp(patch)

                grid[y:y + patch_h, x:x + patch_w] = patch
                idx = idx + 1

    return grid


def make_grid_for_1d_conv(kernels: np.ndarray) -> np.ndarray:
    """Arrange the kernel weights on a grid for visualization purposes.

    Args:
        kernels: The kernel weights of shape (N, 1, 1, IN),
            where N is the number of 2-D convolution kernels with kernel_size=(1, 1) and
            IN is the number of input channels to the 2-D convolution.

    Returns:
        The kernel weights as array for visualization. The array has shape (N * 2 - 1, IN),
            i.e. each second row is a kernel and the other rows are black for separation.
    """
    space = 1
    w = kernels.shape[-1]
    h = kernels.shape[0] + (kernels.shape[0] - 1) * space

    grid = np.zeros([h, w])
    grid_idx = 0
    for kernel_idx in range(kernels.shape[0]):
        grid[grid_idx, :] = kernels[kernel_idx, ...]
        grid_idx += 2

    return grid



import abc
import logging
import time

import numpy as np
import tensorflow as tf
import torch

import pymia.data.definition as defn
import pymia.data.assembler as asmbl
import pymia.deeplearning.data_handler as hdlr
import pymia.deeplearning.model as mdl


class Tester(abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, model: mdl.Model, model_dir: str, log_nth_batch: int):
        """Initializes a new instance of the Tester class.

        The subclasses need to implement following methods:

         - init_subject_assembler
         - predict_batch
         - process_predictions

        Args:
            data_handler: A data handler for the testing dataset.
            model: The model to use for testing.
            model_dir: The directory of the pre-trained model.
            log_nth_batch: Log the progress each n-th predicted batch.
        """
        self.model_dir = model_dir
        self.data_handler = data_handler
        self.model = model

        self.is_model_loaded = False

        self.log_nth_batch = log_nth_batch

    def load(self, model_path: str = None) -> bool:
        """Loads a pre-trained model.

        Args:
            model_path (str): The model's path.
                If `None`, the model will automatically be loaded from the model_dir specified in the constructor.

        Returns:
            True, if the model was loaded; otherwise, False.
        """
        if model_path is not None:
            self.is_model_loaded = self.model.load(model_path)
        else:
            self.is_model_loaded = self.model.load(self.model_dir)

        return self.is_model_loaded

    def predict(self):
        """Predicts subjects based on a pre-trained model.

        This is the main method, which will call the other methods.
        """

        if not self.is_model_loaded:
            self.load()

        start_time = time.time()  # measure the prediction time

        subject_assembler = self.init_subject_assembler()  # todo: not optimal solution since it keeps everything in memory
        self.data_handler.dataset.set_extractor(self.data_handler.extractor_valid)
        self.data_handler.dataset.set_transform(self.data_handler.extraction_transform_valid)
        for batch_idx, batch in enumerate(self.data_handler.loader_test):
            prediction = self.predict_batch(batch_idx, batch)
            subject_assembler.add_batch(prediction, batch)

            if batch_idx % self.log_nth_batch == 0:
                logging.info('Batch {}/{:d}'.format(self._get_batch_index_formatted(batch_idx),
                                                    len(self.data_handler.loader_test)))

        prediction_duration = time.time() - start_time
        logging.info('Prediction time {:.3f} s'.format(prediction_duration))

        self.process_predictions(subject_assembler)

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        """Initializes a subject assembler.

        The subject assembler is used in validate to assemble the predictions on the validation set.

        Returns:
            A subject assembler.
        """
        pass

    @abc.abstractmethod
    def predict_batch(self, idx: int, batch: dict) -> np.ndarray:
        """Predicts a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            The prediction.
        """
        pass

    @abc.abstractmethod
    def process_predictions(self, subject_assembler: asmbl.Assembler):
        """Processes the model's predictions.

        Args:
            subject_assembler: A subject assembler with the assembled predictions of the subjects of the testing set.
        """
        pass

    def _get_batch_index_formatted(self, batch_idx) -> str:
        """Gets the current batch index formatted with leading zeros.

        Returns:
            A string with the formatted batch index.
        """
        return str(batch_idx + 1).zfill(len(str(len(self.data_handler.loader_test))))


class TensorFlowTester(Tester, abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, model: mdl.TensorFlowModel, model_dir: str,
                 session: tf.Session, log_nth_batch: int = 50):
        """Initializes a new instance of the TensorFlowTester class.

        The subclasses need to implement following methods:

         - init_subject_assembler
         - process_predictions

        Args:
            data_handler: A data handler for the testing dataset.
            model: The model to use for testing.
            model_dir: The directory of the pre-trained model.
            session: A TensorFlow session.
            log_nth_batch: Log the progress each n-th predicted batch.
        """
        super().__init__(data_handler, model, model_dir, log_nth_batch)

        self.session = session
        self.model = model  # only required for IntellISense

    def predict_batch(self, idx: int, batch: dict) -> np.ndarray:
        """Predicts a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            The prediction.
        """

        feed_dict = self.batch_to_feed_dict(batch)
        prediction = self.session.run(self.model.network, feed_dict=feed_dict)

        return prediction

    def batch_to_feed_dict(self, batch: dict):
        """Generates the TensorFlow feed dictionary.

        This basic implementation adds x to the feed dictionary. Override this method to add further
        data to the feed dictionary.

        Args:
            batch: The batch from the data loader.

        Returns:
            The feed dictionary.
        """
        feed_dict = {self.model.x_placeholder: np.stack(batch[defn.KEY_IMAGES], axis=0),
                     self.model.is_training_placeholder: False}

        return feed_dict

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        raise NotImplementedError('init_subject_assembler')

    @abc.abstractmethod
    def process_predictions(self, subject_assembler: asmbl.Assembler):
        raise NotImplementedError('process_predictions')


class TorchTester(Tester, abc.ABC):

    def __init__(self, data_handler: hdlr.DataHandler, model: mdl.TorchModel, model_dir: str, log_nth_batch: int = 50):
        """Initializes a new instance of the TorchTester class.

        The subclasses need to implement following methods:

         - init_subject_assembler
         - process_predictions

        Args:
            data_handler: A data handler for the testing dataset.
            model: The model to use for testing.
            model_dir: The directory of the pre-trained model.
            log_nth_batch: Log the progress each n-th predicted batch.
            model: The model to train.
        """
        super().__init__(data_handler, model, model_dir, log_nth_batch)

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model = model  # only required for IntellISense
        self.model.device = self.device
        self.model.network.to(self.device)

    def predict_batch(self, idx: int, batch: dict) -> np.ndarray:
        """Predicts a batch.

        Args:
            idx: The batch index.
            batch: The batch.

        Returns:
            The prediction.
        """

        images = self._get_x(batch)

        self.model.network.eval()

        with torch.no_grad():
            prediction = self.model.inference(images)

        return prediction.cpu().numpy()

    def _get_x(self, batch):
        return batch['images'].to(self.device)

    @abc.abstractmethod
    def init_subject_assembler(self) -> asmbl.Assembler:
        raise NotImplementedError('init_subject_assembler')

    @abc.abstractmethod
    def process_predictions(self, subject_assembler: asmbl.Assembler):
        raise NotImplementedError('process_predictions')



"""The metric module contains a set of evaluation metrics.

The metrics are selected based on the paper of Taha 2016.
Refer to the paper for guidelines how to select appropriate metrics, in-depth description,
and the math.

It is possible to implement your own metrics and use them with the :class:`pymia.evaluation.evaluator.Evaluator`.
Just inherit from :class:`pymia.evaluation.metric.base.IMetric`,
:class:`pymia.evaluation.metric.base.IConfusionMatrixMetric`,
:class:`pymia.evaluation.metric.base.IDistanceMetric`,
:class:`pymia.evaluation.metric.base.ISimpleITKImageMetric` or
:class:`pymia.evaluation.metric.base.INumpyArrayMetric`
and implement the function :func:`pymia.evaluation.metric.base.IMetric.calculate`.
"""
from .regression import (CoefficientOfDetermination, MeanAbsoluteError, MeanSquaredError, RootMeanSquaredError,
                         NormalizedRootMeanSquaredError)
from .segmentation import (Accuracy, AdjustedRandIndex, AreaUnderCurve, AverageDistance, CohenKappaMetric,
                           DiceCoefficient, FalseNegative, FalsePositive, Fallout, FMeasure,
                           GlobalConsistencyError, GroundTruthVolume, HausdorffDistance,
                           InterclassCorrelation, JaccardCoefficient, MahalanobisDistance, MutualInformation, Precision,
                           ProbabilisticDistance, RandIndex, Recall, SegmentationVolume,
                           Sensitivity, Specificity, TrueNegative, TruePositive,
                           VariationOfInformation, VolumeSimilarity)


def get_all_segmentation_metrics():
    """Gets a list with all segmentation metrics.

    Returns:
        list[IMetric]: A list of metrics.
    """
    return get_overlap_metrics() + get_distance_metrics() + get_distance_metrics()


def get_all_regression_metrics():
    """Gets a list with all regression metrics.

    Returns:
        list[IMetric]: A list of metrics.
    """
    return [CoefficientOfDetermination(), MeanAbsoluteError(), MeanSquaredError(), RootMeanSquaredError(),
            NormalizedRootMeanSquaredError()]


def get_overlap_metrics():
    """Gets a list of overlap-based metrics.

    Returns:
        list[IMetric]: A list of metrics.
    """
    return [DiceCoefficient(),
            JaccardCoefficient(),
            AreaUnderCurve(),
            CohenKappaMetric(),
            RandIndex(),
            AdjustedRandIndex(),
            InterclassCorrelation(),
            VolumeSimilarity(),
            MutualInformation()]


def get_distance_metrics():
    """Gets a list of distance-based metrics.

    Returns:
        list[IMetric]: A list of metrics.
    """

    return [HausdorffDistance(),
            AverageDistance(),
            MahalanobisDistance(),
            VariationOfInformation(),
            GlobalConsistencyError(),
            ProbabilisticDistance()]


def get_classical_metrics():
    """Gets a list of classical metrics.

    Returns:
        list[IMetric]: A list of metrics.
    """

    return[Sensitivity(),
           Specificity(),
           Precision(),
           Recall(),
           FMeasure(),
           Accuracy(),
           Fallout(),
           TruePositive(),
           FalsePositive(),
           TrueNegative(),
           FalseNegative(),
           GroundTruthVolume(),
           SegmentationVolume()]
